import logging

from ..utils import enum
from .base import (ADBase, ADComponent as ADCpt, ad_group,
                   EpicsSignalWithRBV as SignalWithRBV)
from ..signal import (EpicsSignalRO, EpicsSignal)
from ..device import DynamicDeviceComponent as DDC

logger = logging.getLogger(__name__)


__all__ = ['CamBase',
           'AdscDetectorCam',
           'Andor3DetectorCam',
           'AndorDetectorCam',
           'BrukerDetectorCam',
           'DexelaDetectorCam',
           'FirewireLinDetectorCam',
           'FirewireWinDetectorCam',
           'GreatEyesDetectorCam',
           'Lambda750kCam',
           'LightFieldDetectorCam',
           'Mar345DetectorCam',
           'MarCCDDetectorCam',
           'PSLDetectorCam',
           'PcoDetectorCam',
           'PcoDetectorIO',
           'PcoDetectorSimIO',
           'PerkinElmerDetectorCam',
           'PilatusDetectorCam',
           'PixiradDetectorCam',
           'PointGreyDetectorCam',
           'ProsilicaDetectorCam',
           'PvcamDetectorCam',
           'RoperDetectorCam',
           'SimDetectorCam',
           'URLDetectorCam',
           ]


class CamBase(ADBase):
    _default_configuration_attrs = (ADBase._default_configuration_attrs +
                                    ('acquire_time', 'acquire_period',
                                     'model', 'num_exposures', 'image_mode',
                                     'manufacturer', 'trigger_mode'))

    ImageMode = enum(SINGLE=0, MULTIPLE=1, CONTINUOUS=2)

    # Shared among all cams and plugins
    array_counter = ADCpt(SignalWithRBV, 'ArrayCounter')
    array_rate = ADCpt(EpicsSignalRO, 'ArrayRate_RBV')
    asyn_io = ADCpt(EpicsSignal, 'AsynIO')

    nd_attributes_file = ADCpt(EpicsSignal, 'NDAttributesFile', string=True)
    pool_alloc_buffers = ADCpt(EpicsSignalRO, 'PoolAllocBuffers')
    pool_free_buffers = ADCpt(EpicsSignalRO, 'PoolFreeBuffers')
    pool_max_buffers = ADCpt(EpicsSignalRO, 'PoolMaxBuffers')
    pool_max_mem = ADCpt(EpicsSignalRO, 'PoolMaxMem')
    pool_used_buffers = ADCpt(EpicsSignalRO, 'PoolUsedBuffers')
    pool_used_mem = ADCpt(EpicsSignalRO, 'PoolUsedMem')
    port_name = ADCpt(EpicsSignalRO, 'PortName_RBV', string=True)

    # Cam-specific
    acquire = ADCpt(SignalWithRBV, 'Acquire')
    acquire_period = ADCpt(SignalWithRBV, 'AcquirePeriod')
    acquire_time = ADCpt(SignalWithRBV, 'AcquireTime')

    array_callbacks = ADCpt(SignalWithRBV, 'ArrayCallbacks')
    array_size = DDC(ad_group(EpicsSignalRO,
                              (('array_size_x', 'ArraySizeX_RBV'),
                               ('array_size_y', 'ArraySizeY_RBV'),
                               ('array_size_z', 'ArraySizeZ_RBV'))),
                     doc='Size of the array in the XYZ dimensions')

    array_size_bytes = ADCpt(EpicsSignalRO, 'ArraySize_RBV')
    bin_x = ADCpt(SignalWithRBV, 'BinX')
    bin_y = ADCpt(SignalWithRBV, 'BinY')
    color_mode = ADCpt(SignalWithRBV, 'ColorMode')
    data_type = ADCpt(SignalWithRBV, 'DataType')
    detector_state = ADCpt(EpicsSignalRO, 'DetectorState_RBV')
    frame_type = ADCpt(SignalWithRBV, 'FrameType')
    gain = ADCpt(SignalWithRBV, 'Gain')

    image_mode = ADCpt(SignalWithRBV, 'ImageMode')
    manufacturer = ADCpt(EpicsSignalRO, 'Manufacturer_RBV')

    max_size = DDC(ad_group(EpicsSignalRO,
                            (('max_size_x', 'MaxSizeX_RBV'),
                             ('max_size_y', 'MaxSizeY_RBV'))),
                   doc='Maximum sensor size in the XY directions')

    min_x = ADCpt(SignalWithRBV, 'MinX')
    min_y = ADCpt(SignalWithRBV, 'MinY')
    model = ADCpt(EpicsSignalRO, 'Model_RBV')

    num_exposures = ADCpt(SignalWithRBV, 'NumExposures')
    num_exposures_counter = ADCpt(EpicsSignalRO, 'NumExposuresCounter_RBV')
    num_images = ADCpt(SignalWithRBV, 'NumImages')
    num_images_counter = ADCpt(EpicsSignalRO, 'NumImagesCounter_RBV')

    read_status = ADCpt(EpicsSignal, 'ReadStatus')
    reverse = DDC(ad_group(SignalWithRBV,
                           (('reverse_x', 'ReverseX'),
                            ('reverse_y', 'ReverseY'))
                           ))

    shutter_close_delay = ADCpt(SignalWithRBV, 'ShutterCloseDelay')
    shutter_close_epics = ADCpt(EpicsSignal, 'ShutterCloseEPICS')
    shutter_control = ADCpt(SignalWithRBV, 'ShutterControl')
    shutter_control_epics = ADCpt(EpicsSignal, 'ShutterControlEPICS')
    shutter_fanout = ADCpt(EpicsSignal, 'ShutterFanout')
    shutter_mode = ADCpt(SignalWithRBV, 'ShutterMode')
    shutter_open_delay = ADCpt(SignalWithRBV, 'ShutterOpenDelay')
    shutter_open_epics = ADCpt(EpicsSignal, 'ShutterOpenEPICS')
    shutter_status_epics = ADCpt(EpicsSignalRO, 'ShutterStatusEPICS_RBV')
    shutter_status = ADCpt(EpicsSignalRO, 'ShutterStatus_RBV')

    size = DDC(ad_group(SignalWithRBV,
                        (('size_x', 'SizeX'),
                         ('size_y', 'SizeY'))
                        ))

    status_message = ADCpt(EpicsSignalRO, 'StatusMessage_RBV', string=True)
    string_from_server = ADCpt(EpicsSignalRO, 'StringFromServer_RBV', string=True)
    string_to_server = ADCpt(EpicsSignalRO, 'StringToServer_RBV', string=True)
    temperature = ADCpt(SignalWithRBV, 'Temperature')
    temperature_actual = ADCpt(EpicsSignal, 'TemperatureActual')
    time_remaining = ADCpt(EpicsSignalRO, 'TimeRemaining_RBV')
    trigger_mode = ADCpt(SignalWithRBV, 'TriggerMode')


class AreaDetectorCam(CamBase):
    pass


class SimDetectorCam(CamBase):
    _html_docs = ['simDetectorDoc.html']
    gain_rgb = DDC(ad_group(SignalWithRBV,
                            (('gain_red', 'GainRed'),
                             ('gain_green', 'GainGreen'),
                             ('gain_blue', 'GainBlue'))),
                   doc='Gain rgb components')
    gain_xy = DDC(ad_group(SignalWithRBV,
                           (('gain_x', 'GainX'),
                            ('gain_y', 'GainY'))),
                  doc='Gain in XY')

    noise = ADCpt(SignalWithRBV, 'Noise')
    peak_num = DDC(ad_group(SignalWithRBV,
                            (('peak_num_x', 'PeakNumX'),
                             ('peak_num_y', 'PeakNumY'))),
                   doc='Peak number in XY')

    peak_start = DDC(ad_group(SignalWithRBV,
                              (('peak_start_x', 'PeakStartX'),
                               ('peak_start_y', 'PeakStartY'))),
                     doc='Peak start in XY')

    peak_step = DDC(ad_group(SignalWithRBV,
                             (('peak_step_x', 'PeakStepX'),
                              ('peak_step_y', 'PeakStepY'))),
                    doc='Peak step in XY')

    peak_variation = ADCpt(SignalWithRBV, 'PeakVariation')
    peak_width = DDC(ad_group(SignalWithRBV,
                              (('peak_width_x', 'PeakWidthX'),
                               ('peak_width_y', 'PeakWidthY'))),
                     doc='Peak width in XY')

    reset = ADCpt(SignalWithRBV, 'Reset')
    sim_mode = ADCpt(SignalWithRBV, 'SimMode')


class AdscDetectorCam(CamBase):
    _html_docs = ['adscDoc.html']
    adsc_2theta = ADCpt(SignalWithRBV, 'ADSC2Theta')
    adsc_adc = ADCpt(SignalWithRBV, 'ADSCAdc')
    adsc_axis = ADCpt(SignalWithRBV, 'ADSCAxis')
    adsc_beam_x = ADCpt(SignalWithRBV, 'ADSCBeamX')
    adsc_beam_y = ADCpt(SignalWithRBV, 'ADSCBeamY')
    adsc_dezingr = ADCpt(SignalWithRBV, 'ADSCDezingr')
    adsc_distance = ADCpt(SignalWithRBV, 'ADSCDistnce')
    adsc_im_width = ADCpt(SignalWithRBV, 'ADSCImWidth')
    adsc_im_xform = ADCpt(SignalWithRBV, 'ADSCImXform')
    adsc_kappa = ADCpt(SignalWithRBV, 'ADSCKappa')
    adsc_last_error = ADCpt(EpicsSignal, 'ADSCLastError')
    adsc_last_image = ADCpt(EpicsSignal, 'ADSCLastImage')
    adsc_omega = ADCpt(SignalWithRBV, 'ADSCOmega')
    adsc_phi = ADCpt(SignalWithRBV, 'ADSCPhi')
    adsc_raw = ADCpt(SignalWithRBV, 'ADSCRaw')
    adsc_read_conditn = ADCpt(EpicsSignal, 'ADSCReadConditn')
    adsc_reus_drk = ADCpt(SignalWithRBV, 'ADSCReusDrk')
    adsc_soft_reset = ADCpt(EpicsSignal, 'ADSCSoftReset')
    adsc_state = ADCpt(EpicsSignal, 'ADSCState')
    adsc_status = ADCpt(EpicsSignal, 'ADSCStatus')
    adsc_stp_ex_retry_count = ADCpt(EpicsSignal, 'ADSCStpExRtryCt')
    adsc_str_drks = ADCpt(SignalWithRBV, 'ADSCStrDrks')
    adsc_wavelen = ADCpt(SignalWithRBV, 'ADSCWavelen')

    bin_x_changed = ADCpt(EpicsSignal, 'BinXChanged')
    bin_y_changed = ADCpt(EpicsSignal, 'BinYChanged')
    ext_trig_ctl = ADCpt(EpicsSignal, 'ExSwTrCtl')
    ext_trig_ctl_rsp = ADCpt(EpicsSignal, 'ExSwTrCtlRsp')
    ext_trig_ok_to_exp = ADCpt(EpicsSignal, 'ExSwTrOkToExp')


class AndorDetectorCam(CamBase):
    _html_docs = ['andorDoc.html']
    andor_adc_speed = ADCpt(SignalWithRBV, 'AndorADCSpeed')
    andor_accumulate_period = ADCpt(SignalWithRBV, 'AndorAccumulatePeriod')
    andor_cooler = ADCpt(SignalWithRBV, 'AndorCooler')
    andor_message = ADCpt(EpicsSignalRO, 'AndorMessage_RBV')
    andor_pre_amp_gain = ADCpt(SignalWithRBV, 'AndorPreAmpGain')
    andor_shutter_ex_ttl = ADCpt(EpicsSignal, 'AndorShutterExTTL')
    andor_shutter_mode = ADCpt(EpicsSignal, 'AndorShutterMode')
    andor_temp_status = ADCpt(EpicsSignalRO, 'AndorTempStatus_RBV')
    file_format = ADCpt(SignalWithRBV, 'FileFormat')
    pal_file_path = ADCpt(SignalWithRBV, 'PALFilePath')


class Andor3DetectorCam(CamBase):
    _html_docs = ['andor3Doc.html']
    a3_binning = ADCpt(SignalWithRBV, 'A3Binning')
    a3_shutter_mode = ADCpt(SignalWithRBV, 'A3ShutterMode')
    controller_id = ADCpt(EpicsSignal, 'ControllerID')
    fan_speed = ADCpt(SignalWithRBV, 'FanSpeed')
    firmware_version = ADCpt(EpicsSignal, 'FirmwareVersion')
    frame_rate = ADCpt(SignalWithRBV, 'FrameRate')
    full_aoic_ontrol = ADCpt(EpicsSignal, 'FullAOIControl')
    noise_filter = ADCpt(SignalWithRBV, 'NoiseFilter')
    overlap = ADCpt(SignalWithRBV, 'Overlap')
    pixel_encoding = ADCpt(SignalWithRBV, 'PixelEncoding')
    pre_amp_gain = ADCpt(SignalWithRBV, 'PreAmpGain')
    readout_rate = ADCpt(SignalWithRBV, 'ReadoutRate')
    readout_time = ADCpt(EpicsSignal, 'ReadoutTime')
    sensor_cooling = ADCpt(SignalWithRBV, 'SensorCooling')
    serial_number = ADCpt(EpicsSignal, 'SerialNumber')
    software_trigger = ADCpt(EpicsSignal, 'SoftwareTrigger')
    software_version = ADCpt(EpicsSignal, 'SoftwareVersion')
    temp_control = ADCpt(SignalWithRBV, 'TempControl')
    temp_status = ADCpt(EpicsSignalRO, 'TempStatus_RBV')
    transfer_rate = ADCpt(EpicsSignal, 'TransferRate')


class BrukerDetectorCam(CamBase):
    _html_docs = ['BrukerDoc.html']
    bis_asyn = ADCpt(EpicsSignal, 'BISAsyn')
    bis_status = ADCpt(EpicsSignal, 'BISStatus')
    file_format = ADCpt(SignalWithRBV, 'FileFormat')
    num_darks = ADCpt(SignalWithRBV, 'NumDarks')
    read_sfrm_timeout = ADCpt(EpicsSignal, 'ReadSFRMTimeout')


class DexelaDetectorCam(CamBase):
    acquire_gain = ADCpt(EpicsSignal, 'DEXAcquireGain')
    acquire_offset = ADCpt(EpicsSignal, 'DEXAcquireOffset')
    binning_mode = ADCpt(SignalWithRBV, 'DEXBinningMode')
    corrections_dir = ADCpt(EpicsSignal, 'DEXCorrectionsDir', string=True)
    current_gain_frame = ADCpt(EpicsSignal, 'DEXCurrentGainFrame')
    current_offset_frame = ADCpt(EpicsSignal, 'DEXCurrentOffsetFrame')
    defect_map_available = ADCpt(EpicsSignal, 'DEXDefectMapAvailable')
    defect_map_file = ADCpt(EpicsSignal, 'DEXDefectMapFile', string=True)
    full_well_mode = ADCpt(SignalWithRBV, 'DEXFullWellMode')
    gain_available = ADCpt(EpicsSignal, 'DEXGainAvailable')
    gain_file = ADCpt(EpicsSignal, 'DEXGainFile', string=True)
    load_defect_map_file = ADCpt(EpicsSignal, 'DEXLoadDefectMapFile')
    load_gain_file = ADCpt(EpicsSignal, 'DEXLoadGainFile')
    load_offset_file = ADCpt(EpicsSignal, 'DEXLoadOffsetFile')
    num_gain_frames = ADCpt(EpicsSignal, 'DEXNumGainFrames')
    num_offset_frames = ADCpt(EpicsSignal, 'DEXNumOffsetFrames')
    offset_available = ADCpt(EpicsSignal, 'DEXOffsetAvailable')
    offset_constant = ADCpt(SignalWithRBV, 'DEXOffsetConstant')
    offset_file = ADCpt(EpicsSignal, 'DEXOffsetFile', string=True)
    save_gain_file = ADCpt(EpicsSignal, 'DEXSaveGainFile')
    save_offset_file = ADCpt(EpicsSignal, 'DEXSaveOffsetFile')
    serial_number = ADCpt(EpicsSignal, 'DEXSerialNumber')
    software_trigger = ADCpt(EpicsSignal, 'DEXSoftwareTrigger')
    use_defect_map = ADCpt(EpicsSignal, 'DEXUseDefectMap')
    use_gain = ADCpt(EpicsSignal, 'DEXUseGain')
    use_offset = ADCpt(EpicsSignal, 'DEXUseOffset')


class FirewireLinDetectorCam(CamBase):
    _html_docs = []

    bandwidth = ADCpt(EpicsSignal, 'BANDWIDTH')
    fanout_disable = ADCpt(EpicsSignal, 'FanoutDis')
    framerate_max = ADCpt(SignalWithRBV, 'FR')
    is_fixed_mode = ADCpt(EpicsSignal, 'IsFixedMode')
    video_mode = ADCpt(EpicsSignal, 'VIDEOMODE')


class FirewireWinDetectorCam(CamBase):
    _html_docs = ['FirewireWinDoc.html']
    colorcode = ADCpt(SignalWithRBV, 'COLORCODE')
    current_colorcode = ADCpt(EpicsSignal, 'CURRENT_COLORCODE')
    current_format = ADCpt(EpicsSignal, 'CURRENT_FORMAT')
    current_mode = ADCpt(EpicsSignal, 'CURRENT_MODE')
    current_rate = ADCpt(EpicsSignal, 'CURRENT_RATE')
    dropped_frames = ADCpt(SignalWithRBV, 'DROPPED_FRAMES')
    format_ = ADCpt(SignalWithRBV, 'FORMAT')
    frame_rate = ADCpt(SignalWithRBV, 'FR')
    mode = ADCpt(SignalWithRBV, 'MODE')
    readout_time = ADCpt(SignalWithRBV, 'READOUT_TIME')


class GreatEyesDetectorCam(CamBase):
    _html_docs = []
    adc_speed = ADCpt(SignalWithRBV, 'GreatEyesAdcSpeed')
    capacity = ADCpt(SignalWithRBV, 'GreatEyesCapacity')
    enable_cooling = ADCpt(SignalWithRBV, 'GreatEyesEnableCooling')
    gain = ADCpt(SignalWithRBV, 'GreatEyesGain')
    hot_side_temp = ADCpt(EpicsSignal, 'GreatEyesHotSideTemp')
    readout_dir = ADCpt(SignalWithRBV, 'GreatEyesReadoutDir')
    sync = ADCpt(SignalWithRBV, 'GreatEyesSync')


class Lambda750kCam(CamBase):
    """
    support for X-Spectrum Lambda 750K detector

    https://x-spectrum.de/products/lambda-350k750k/
    """
    _html_docs = ['Lambda750kCam.html']

    config_file_path = ADCpt(EpicsSignal, 'ConfigFilePath')
    firmware_version = ADCpt(EpicsSignalRO, 'FirmwareVersion_RBV')
    operating_mode = ADCpt(SignalWithRBV, 'OperatingMode')
    serial_number = ADCpt(EpicsSignalRO, 'SerialNumber_RBV')
    temperature = ADCpt(SignalWithRBV, 'Temperature')


class LightFieldDetectorCam(CamBase):
    _html_docs = ['LightFieldDoc.html']

    aux_delay = ADCpt(SignalWithRBV, 'LFAuxDelay')
    aux_width = ADCpt(SignalWithRBV, 'LFAuxWidth')
    background_enable = ADCpt(SignalWithRBV, 'LFBackgroundEnable')
    background_file = ADCpt(SignalWithRBV, 'LFBackgroundFile')
    background_full_file = ADCpt(EpicsSignalRO, 'LFBackgroundFullFile_RBV')
    background_path = ADCpt(SignalWithRBV, 'LFBackgroundPath')
    entrance_width = ADCpt(SignalWithRBV, 'LFEntranceWidth')
    exit_port = ADCpt(SignalWithRBV, 'LFExitPort')
    experiment_name = ADCpt(SignalWithRBV, 'LFExperimentName')
    file_name = ADCpt(EpicsSignalRO, 'LFFileName_RBV')
    file_path = ADCpt(EpicsSignalRO, 'LFFilePath_RBV')
    lf_gain = ADCpt(SignalWithRBV, 'LFGain')
    gating_mode = ADCpt(SignalWithRBV, 'LFGatingMode')
    grating = ADCpt(SignalWithRBV, 'LFGrating')
    grating_wavelength = ADCpt(SignalWithRBV, 'LFGratingWavelength')
    image_mode = ADCpt(SignalWithRBV, 'ImageMode')
    intensifier_enable = ADCpt(SignalWithRBV, 'LFIntensifierEnable')
    intensifier_gain = ADCpt(SignalWithRBV, 'LFIntensifierGain')
    num_accumulations = ADCpt(SignalWithRBV, 'NumAccumulations')
    ready_to_run = ADCpt(EpicsSignal, 'LFReadyToRun')
    rep_gate_delay = ADCpt(SignalWithRBV, 'LFRepGateDelay')
    rep_gate_width = ADCpt(SignalWithRBV, 'LFRepGateWidth')
    seq_end_gate_delay = ADCpt(SignalWithRBV, 'LFSeqEndGateDelay')
    seq_end_gate_width = ADCpt(SignalWithRBV, 'LFSeqEndGateWidth')
    seq_start_gate_delay = ADCpt(SignalWithRBV, 'LFSeqStartGateDelay')
    seq_start_gate_width = ADCpt(SignalWithRBV, 'LFSeqStartGateWidth')
    lf_shutter_mode = ADCpt(SignalWithRBV, 'LFShutterMode')
    sync_master2_delay = ADCpt(SignalWithRBV, 'LFSyncMaster2Delay')
    sync_master_enable = ADCpt(SignalWithRBV, 'LFSyncMasterEnable')
    trigger_frequency = ADCpt(SignalWithRBV, 'LFTriggerFrequency')
    update_experiments = ADCpt(EpicsSignal, 'LFUpdateExperiments')


class Mar345DetectorCam(CamBase):
    _html_docs = ['Mar345Doc.html']
    abort = ADCpt(SignalWithRBV, 'Abort')
    change_mode = ADCpt(SignalWithRBV, 'ChangeMode')
    erase = ADCpt(SignalWithRBV, 'Erase')
    erase_mode = ADCpt(SignalWithRBV, 'EraseMode')
    file_format = ADCpt(SignalWithRBV, 'FileFormat')
    num_erase = ADCpt(SignalWithRBV, 'NumErase')
    num_erased = ADCpt(EpicsSignalRO, 'NumErased_RBV')
    scan_resolution = ADCpt(SignalWithRBV, 'ScanResolution')
    scan_size = ADCpt(SignalWithRBV, 'ScanSize')
    mar_server_asyn = ADCpt(EpicsSignal, 'marServerAsyn')


class MarCCDDetectorCam(CamBase):
    _html_docs = ['MarCCDDoc.html']
    beam_x = ADCpt(EpicsSignal, 'BeamX')
    beam_y = ADCpt(EpicsSignal, 'BeamY')
    dataset_comments = ADCpt(EpicsSignal, 'DatasetComments')
    detector_distance = ADCpt(EpicsSignal, 'DetectorDistance')
    file_comments = ADCpt(EpicsSignal, 'FileComments')
    file_format = ADCpt(SignalWithRBV, 'FileFormat')
    frame_shift = ADCpt(SignalWithRBV, 'FrameShift')
    mar_acquire_status = ADCpt(EpicsSignalRO, 'MarAcquireStatus_RBV')
    mar_correct_status = ADCpt(EpicsSignalRO, 'MarCorrectStatus_RBV')
    mar_dezinger_status = ADCpt(EpicsSignalRO, 'MarDezingerStatus_RBV')
    mar_readout_status = ADCpt(EpicsSignalRO, 'MarReadoutStatus_RBV')
    mar_state = ADCpt(EpicsSignalRO, 'MarState_RBV')
    mar_status = ADCpt(EpicsSignalRO, 'MarStatus_RBV')
    mar_writing_status = ADCpt(EpicsSignalRO, 'MarWritingStatus_RBV')
    overlap_mode = ADCpt(SignalWithRBV, 'OverlapMode')
    read_tiff_timeout = ADCpt(EpicsSignal, 'ReadTiffTimeout')
    rotation_axis = ADCpt(EpicsSignal, 'RotationAxis')
    rotation_range = ADCpt(EpicsSignal, 'RotationRange')
    stability = ADCpt(SignalWithRBV, 'Stability')
    start_phi = ADCpt(EpicsSignal, 'StartPhi')
    two_theta = ADCpt(EpicsSignal, 'TwoTheta')
    wavelength = ADCpt(EpicsSignal, 'Wavelength')
    mar_server_asyn = ADCpt(EpicsSignal, 'marServerAsyn')


class PcoDetectorCam(CamBase):
    _html_docs = ['']
    adc_mode = ADCpt(SignalWithRBV, 'ADC_MODE')
    arm_mode = ADCpt(SignalWithRBV, 'ARM_MODE')
    bit_alignment = ADCpt(SignalWithRBV, 'BIT_ALIGNMENT')
    camera_setup = ADCpt(SignalWithRBV, 'CAMERA_SETUP')
    cam_ram_use = ADCpt(EpicsSignalRO, 'CAM_RAM_USE_RBV')
    delay_time = ADCpt(SignalWithRBV, 'DELAY_TIME')
    elec_temp = ADCpt(EpicsSignalRO, 'ELEC_TEMP_RBV')
    exposure_base = ADCpt(EpicsSignalRO, 'EXPOSUREBASE_RBV')
    pco_acquire_mode = ADCpt(SignalWithRBV, 'ACQUIRE_MODE')
    pco_image_number = ADCpt(EpicsSignalRO, 'IMAGE_NUMBER_RBV')
    pix_rate = ADCpt(SignalWithRBV, 'PIX_RATE')
    power_temp = ADCpt(EpicsSignalRO, 'POWER_TEMP_RBV')
    recorder_mode = ADCpt(SignalWithRBV, 'RECORDER_MODE')
    storage_mode = ADCpt(SignalWithRBV, 'STORAGE_MODE')
    timestamp_mode = ADCpt(SignalWithRBV, 'TIMESTAMP_MODE')


class PcoDetectorIO(ADBase):
    _html_docs = ['']
    busy = ADCpt(EpicsSignal, 'DIO:BUSY')
    capture = ADCpt(EpicsSignal, 'DIO:CAPTURE')
    exposing = ADCpt(EpicsSignal, 'DIO:EXPOSING')
    ready = ADCpt(EpicsSignal, 'DIO:READY')
    trig = ADCpt(EpicsSignal, 'DIO:TRIG')
    trig_when_ready = ADCpt(EpicsSignal, 'DIO:TrigWhenReady')


class PcoDetectorSimIO(ADBase):
    _html_docs = ['']
    busy = ADCpt(EpicsSignal, 'SIM:BUSY')
    dfan = ADCpt(EpicsSignal, 'SIM:Dfan')
    exposing = ADCpt(EpicsSignal, 'SIM:EXPOSING')
    set_busy = ADCpt(EpicsSignal, 'SIM:SetBusy')
    set_exp = ADCpt(EpicsSignal, 'SIM:SetExp')
    set_state = ADCpt(EpicsSignal, 'SIM:SetState')
    trig = ADCpt(EpicsSignal, 'SIM:TRIG')


class PerkinElmerDetectorCam(CamBase):
    _html_docs = ['PerkinElmerDoc.html']
    pe_acquire_gain = ADCpt(EpicsSignal, 'PEAcquireGain')
    pe_acquire_offset = ADCpt(EpicsSignal, 'PEAcquireOffset')
    pe_corrections_dir = ADCpt(EpicsSignal, 'PECorrectionsDir')
    pe_current_gain_frame = ADCpt(EpicsSignal, 'PECurrentGainFrame')
    pe_current_offset_frame = ADCpt(EpicsSignal, 'PECurrentOffsetFrame')
    pe_dwell_time = ADCpt(SignalWithRBV, 'PEDwellTime')
    pe_frame_buff_index = ADCpt(EpicsSignal, 'PEFrameBuffIndex')
    pe_gain = ADCpt(SignalWithRBV, 'PEGain')
    pe_gain_available = ADCpt(EpicsSignal, 'PEGainAvailable')
    pe_gain_file = ADCpt(EpicsSignal, 'PEGainFile')
    pe_image_number = ADCpt(EpicsSignal, 'PEImageNumber')
    pe_initialize = ADCpt(EpicsSignal, 'PEInitialize')
    pe_load_gain_file = ADCpt(EpicsSignal, 'PELoadGainFile')
    pe_load_pixel_correction = ADCpt(EpicsSignal, 'PELoadPixelCorrection')
    pe_num_frame_buffers = ADCpt(SignalWithRBV, 'PENumFrameBuffers')
    pe_num_frames_to_skip = ADCpt(SignalWithRBV, 'PENumFramesToSkip')
    pe_num_gain_frames = ADCpt(EpicsSignal, 'PENumGainFrames')
    pe_num_offset_frames = ADCpt(EpicsSignal, 'PENumOffsetFrames')
    pe_offset_available = ADCpt(EpicsSignal, 'PEOffsetAvailable')
    pe_pixel_correction_available = ADCpt(EpicsSignal,
                                          'PEPixelCorrectionAvailable')
    pe_pixel_correction_file = ADCpt(EpicsSignal, 'PEPixelCorrectionFile')
    pe_save_gain_file = ADCpt(EpicsSignal, 'PESaveGainFile')
    pe_skip_frames = ADCpt(SignalWithRBV, 'PESkipFrames')
    pe_sync_time = ADCpt(SignalWithRBV, 'PESyncTime')
    pe_system_id = ADCpt(EpicsSignal, 'PESystemID')
    pe_trigger = ADCpt(EpicsSignal, 'PETrigger')
    pe_use_gain = ADCpt(EpicsSignal, 'PEUseGain')
    pe_use_offset = ADCpt(EpicsSignal, 'PEUseOffset')
    pe_use_pixel_correction = ADCpt(EpicsSignal, 'PEUsePixelCorrection')


class PSLDetectorCam(CamBase):
    _html_docs = ['PSLDoc.html']
    file_format = ADCpt(SignalWithRBV, 'FileFormat')
    tiff_comment = ADCpt(SignalWithRBV, 'TIFFComment')


class PilatusDetectorCam(CamBase):
    _html_docs = ['pilatusDoc.html']
    alpha = ADCpt(EpicsSignal, 'Alpha')
    angle_incr = ADCpt(EpicsSignal, 'AngleIncr')
    armed = ADCpt(EpicsSignal, 'Armed')
    bad_pixel_file = ADCpt(EpicsSignal, 'BadPixelFile')
    beam_x = ADCpt(EpicsSignal, 'BeamX')
    beam_y = ADCpt(EpicsSignal, 'BeamY')
    camserver_asyn = ADCpt(EpicsSignal, 'CamserverAsyn')
    cbf_template_file = ADCpt(EpicsSignal, 'CbfTemplateFile')
    chi = ADCpt(EpicsSignal, 'Chi')
    delay_time = ADCpt(SignalWithRBV, 'DelayTime')
    det_2theta = ADCpt(EpicsSignal, 'Det2theta')
    det_dist = ADCpt(EpicsSignal, 'DetDist')
    det_v_offset = ADCpt(EpicsSignal, 'DetVOffset')
    energy_high = ADCpt(EpicsSignal, 'EnergyHigh')
    energy_low = ADCpt(EpicsSignal, 'EnergyLow')
    file_format = ADCpt(SignalWithRBV, 'FileFormat')
    filter_transm = ADCpt(EpicsSignal, 'FilterTransm')
    flat_field_file = ADCpt(EpicsSignal, 'FlatFieldFile')
    flat_field_valid = ADCpt(EpicsSignal, 'FlatFieldValid')
    flux = ADCpt(EpicsSignal, 'Flux')
    gain_menu = ADCpt(EpicsSignal, 'GainMenu')
    gap_fill = ADCpt(SignalWithRBV, 'GapFill')
    header_string = ADCpt(EpicsSignal, 'HeaderString')
    humid0 = ADCpt(EpicsSignalRO, 'Humid0_RBV')
    humid1 = ADCpt(EpicsSignalRO, 'Humid1_RBV')
    humid2 = ADCpt(EpicsSignalRO, 'Humid2_RBV')
    image_file_tmot = ADCpt(EpicsSignal, 'ImageFileTmot')
    kappa = ADCpt(EpicsSignal, 'Kappa')
    min_flat_field = ADCpt(SignalWithRBV, 'MinFlatField')
    num_bad_pixels = ADCpt(EpicsSignal, 'NumBadPixels')
    num_oscill = ADCpt(EpicsSignal, 'NumOscill')
    oscill_axis = ADCpt(EpicsSignal, 'OscillAxis')
    phi = ADCpt(EpicsSignal, 'Phi')
    pixel_cut_off = ADCpt(EpicsSignalRO, 'PixelCutOff_RBV')
    polarization = ADCpt(EpicsSignal, 'Polarization')
    start_angle = ADCpt(EpicsSignal, 'StartAngle')
    tvx_version = ADCpt(EpicsSignalRO, 'TVXVersion_RBV')
    temp0 = ADCpt(EpicsSignalRO, 'Temp0_RBV')
    temp1 = ADCpt(EpicsSignalRO, 'Temp1_RBV')
    temp2 = ADCpt(EpicsSignalRO, 'Temp2_RBV')
    threshold_apply = ADCpt(EpicsSignal, 'ThresholdApply')
    threshold_auto_apply = ADCpt(SignalWithRBV, 'ThresholdAutoApply')
    threshold_energy = ADCpt(SignalWithRBV, 'ThresholdEnergy')
    wavelength = ADCpt(EpicsSignal, 'Wavelength')


class PixiradDetectorCam(CamBase):
    _html_docs = ['PixiradDoc.html']

    auto_calibrate = ADCpt(EpicsSignal, 'AutoCalibrate')
    humidity_box = ADCpt(EpicsSignalRO, 'BoxHumidity_RBV')
    colors_collected = ADCpt(EpicsSignalRO, 'ColorsCollected_RBV')
    cooling_state = ADCpt(SignalWithRBV, 'CoolingState')
    cooling_status = ADCpt(EpicsSignalRO, 'CoolingStatus_RBV')
    dew_point = ADCpt(EpicsSignalRO, 'DewPoint_RBV')
    frame_type = ADCpt(SignalWithRBV, 'FrameType')
    hv_actual = ADCpt(EpicsSignalRO, 'HVActual_RBV')
    hv_current = ADCpt(EpicsSignalRO, 'HVCurrent_RBV')
    hv_mode = ADCpt(SignalWithRBV, 'HVMode')
    hv_state = ADCpt(SignalWithRBV, 'HVState')
    hv_value = ADCpt(SignalWithRBV, 'HVValue')
    peltier_power = ADCpt(EpicsSignalRO, 'PeltierPower_RBV')
    sync_in_polarity = ADCpt(SignalWithRBV, 'SyncInPolarity')
    sync_out_function = ADCpt(SignalWithRBV, 'SyncOutFunction')
    sync_out_polarity = ADCpt(SignalWithRBV, 'SyncOutPolarity')
    system_reset = ADCpt(EpicsSignal, 'SystemReset')

    temperature = ADCpt(SignalWithRBV, 'Temperature')
    temperature_actual = ADCpt(EpicsSignal, 'TemperatureActual')
    temperature_box = ADCpt(EpicsSignalRO, 'BoxTemperature_RBV')
    temperature_hot = ADCpt(EpicsSignalRO, 'HotTemperature_RBV')

    threshold_1_actual = ADCpt(EpicsSignalRO, 'ThresholdActual1_RBV')
    threshold_2_actual = ADCpt(EpicsSignalRO, 'ThresholdActual2_RBV')
    threshold_3_actual = ADCpt(EpicsSignalRO, 'ThresholdActual3_RBV')
    threshold_4_actual = ADCpt(EpicsSignalRO, 'ThresholdActual4_RBV')
    thresholds_actual = DDC(ad_group(EpicsSignalRO,
                                     (('threshold_1', 'ThresholdActual1_RBV'),
                                      ('threshold_2', 'ThresholdActual2_RBV'),
                                      ('threshold_3', 'ThresholdActual3_RBV'),
                                      ('threshold_4', 'ThresholdActual4_RBV'),
                                      )),
                            doc='Actual thresholds')

    threshold_1 = ADCpt(SignalWithRBV, 'Threshold1')
    threshold_2 = ADCpt(SignalWithRBV, 'Threshold2')
    threshold_3 = ADCpt(SignalWithRBV, 'Threshold3')
    threshold_4 = ADCpt(SignalWithRBV, 'Threshold4')
    thresholds = DDC(ad_group(SignalWithRBV,
                              (('threshold_1', 'Threshold1'),
                               ('threshold_2', 'Threshold2'),
                               ('threshold_3', 'Threshold3'),
                               ('threshold_4', 'Threshold4'),
                               )),
                     doc='Thresholds')

    udp_buffers_free = ADCpt(EpicsSignalRO, 'UDPBuffersFree_RBV')
    udp_buffers_max = ADCpt(EpicsSignalRO, 'UDPBuffersMax_RBV')
    udp_buffers_read = ADCpt(EpicsSignalRO, 'UDPBuffersRead_RBV')
    udp_speed = ADCpt(EpicsSignalRO, 'UDPSpeed_RBV')


class PointGreyDetectorCam(CamBase):
    _html_docs = ['PointGreyDoc.html']

    bandwidth = ADCpt(EpicsSignal, 'Bandwidth')
    binning_mode = ADCpt(SignalWithRBV, 'BinningMode')
    convert_pixel_format = ADCpt(SignalWithRBV, 'ConvertPixelFormat')
    corrupt_frames = ADCpt(EpicsSignalRO, 'CorruptFrames_RBV')
    driver_dropped = ADCpt(EpicsSignalRO, 'DriverDropped_RBV')
    dropped_frames = ADCpt(EpicsSignalRO, 'DroppedFrames_RBV')
    firmware_version = ADCpt(EpicsSignal, 'FirmwareVersion')
    format7_mode = ADCpt(SignalWithRBV, 'Format7Mode')
    frame_rate = ADCpt(SignalWithRBV, 'FrameRate')
    max_packet_size = ADCpt(EpicsSignal, 'MaxPacketSize')
    packet_delay_actual = ADCpt(EpicsSignal, 'PacketDelayActual')
    packet_delay = ADCpt(SignalWithRBV, 'PacketDelay')
    packet_size_actual = ADCpt(EpicsSignal, 'PacketSizeActual')
    packet_size = ADCpt(SignalWithRBV, 'PacketSize')
    pixel_format = ADCpt(SignalWithRBV, 'PixelFormat')
    read_status = ADCpt(EpicsSignal, 'ReadStatus')
    serial_number = ADCpt(EpicsSignal, 'SerialNumber')
    skip_frames = ADCpt(SignalWithRBV, 'SkipFrames')
    software_trigger = ADCpt(EpicsSignal, 'SoftwareTrigger')
    software_version = ADCpt(EpicsSignal, 'SoftwareVersion')
    strobe_delay = ADCpt(SignalWithRBV, 'StrobeDelay')
    strobe_duration = ADCpt(SignalWithRBV, 'StrobeDuration')
    strobe_enable = ADCpt(SignalWithRBV, 'StrobeEnable')
    strobe_polarity = ADCpt(SignalWithRBV, 'StrobePolarity')
    strobe_source = ADCpt(SignalWithRBV, 'StrobeSource')
    time_stamp_mode = ADCpt(SignalWithRBV, 'TimeStampMode')
    transmit_failed = ADCpt(EpicsSignalRO, 'TransmitFailed_RBV')
    trigger_polarity = ADCpt(SignalWithRBV, 'TriggerPolarity')
    trigger_source = ADCpt(SignalWithRBV, 'TriggerSource')
    video_mode = ADCpt(SignalWithRBV, 'VideoMode')


class ProsilicaDetectorCam(CamBase):
    _html_docs = ['prosilicaDoc.html']
    ps_bad_frame_counter = ADCpt(EpicsSignalRO, 'PSBadFrameCounter_RBV')
    ps_byte_rate = ADCpt(SignalWithRBV, 'PSByteRate')
    ps_driver_type = ADCpt(EpicsSignalRO, 'PSDriverType_RBV')
    ps_filter_version = ADCpt(EpicsSignalRO, 'PSFilterVersion_RBV')
    ps_frame_rate = ADCpt(EpicsSignalRO, 'PSFrameRate_RBV')
    ps_frames_completed = ADCpt(EpicsSignalRO, 'PSFramesCompleted_RBV')
    ps_frames_dropped = ADCpt(EpicsSignalRO, 'PSFramesDropped_RBV')
    ps_packet_size = ADCpt(EpicsSignalRO, 'PSPacketSize_RBV')
    ps_packets_erroneous = ADCpt(EpicsSignalRO, 'PSPacketsErroneous_RBV')
    ps_packets_missed = ADCpt(EpicsSignalRO, 'PSPacketsMissed_RBV')
    ps_packets_received = ADCpt(EpicsSignalRO, 'PSPacketsReceived_RBV')
    ps_packets_requested = ADCpt(EpicsSignalRO, 'PSPacketsRequested_RBV')
    ps_packets_resent = ADCpt(EpicsSignalRO, 'PSPacketsResent_RBV')
    ps_read_statistics = ADCpt(EpicsSignal, 'PSReadStatistics')
    ps_reset_timer = ADCpt(EpicsSignal, 'PSResetTimer')
    ps_timestamp_type = ADCpt(SignalWithRBV, 'PSTimestampType')
    strobe1_ctl_duration = ADCpt(SignalWithRBV, 'Strobe1CtlDuration')
    strobe1_delay = ADCpt(SignalWithRBV, 'Strobe1Delay')
    strobe1_duration = ADCpt(SignalWithRBV, 'Strobe1Duration')
    strobe1_mode = ADCpt(SignalWithRBV, 'Strobe1Mode')
    sync_in1_level = ADCpt(EpicsSignalRO, 'SyncIn1Level_RBV')
    sync_in2_level = ADCpt(EpicsSignalRO, 'SyncIn2Level_RBV')
    sync_out1_invert = ADCpt(SignalWithRBV, 'SyncOut1Invert')
    sync_out1_level = ADCpt(SignalWithRBV, 'SyncOut1Level')
    sync_out1_mode = ADCpt(SignalWithRBV, 'SyncOut1Mode')
    sync_out2_invert = ADCpt(SignalWithRBV, 'SyncOut2Invert')
    sync_out2_level = ADCpt(SignalWithRBV, 'SyncOut2Level')
    sync_out2_mode = ADCpt(SignalWithRBV, 'SyncOut2Mode')
    sync_out3_invert = ADCpt(SignalWithRBV, 'SyncOut3Invert')
    sync_out3_level = ADCpt(SignalWithRBV, 'SyncOut3Level')
    sync_out3_mode = ADCpt(SignalWithRBV, 'SyncOut3Mode')
    trigger_delay = ADCpt(SignalWithRBV, 'TriggerDelay')
    trigger_event = ADCpt(SignalWithRBV, 'TriggerEvent')
    trigger_overlap = ADCpt(SignalWithRBV, 'TriggerOverlap')
    trigger_software = ADCpt(EpicsSignal, 'TriggerSoftware')


class PvcamDetectorCam(CamBase):
    _html_docs = ['pvcamDoc.html']
    bit_depth = ADCpt(EpicsSignalRO, 'BitDepth_RBV')
    camera_firmware_vers = ADCpt(EpicsSignalRO, 'CameraFirmwareVers_RBV')
    chip_height = ADCpt(EpicsSignalRO, 'ChipHeight_RBV')
    chip_name = ADCpt(EpicsSignalRO, 'ChipName_RBV')
    chip_width = ADCpt(EpicsSignalRO, 'ChipWidth_RBV')
    close_delay = ADCpt(SignalWithRBV, 'CloseDelay')
    detector_mode = ADCpt(SignalWithRBV, 'DetectorMode')
    detector_selected = ADCpt(SignalWithRBV, 'DetectorSelected')
    dev_drv_vers = ADCpt(EpicsSignalRO, 'DevDrvVers_RBV')
    frame_transfer_capable = ADCpt(EpicsSignalRO, 'FrameTransferCapable_RBV')
    full_well_capacity = ADCpt(EpicsSignalRO, 'FullWellCapacity_RBV')
    gain_index = ADCpt(SignalWithRBV, 'GainIndex')
    head_ser_num = ADCpt(EpicsSignalRO, 'HeadSerNum_RBV')
    initialize = ADCpt(SignalWithRBV, 'Initialize')
    max_gain_index = ADCpt(EpicsSignalRO, 'MaxGainIndex_RBV')
    max_set_temperature = ADCpt(EpicsSignal, 'MaxSetTemperature')
    max_shutter_close_delay = ADCpt(EpicsSignalRO, 'MaxShutterCloseDelay_RBV')
    max_shutter_open_delay = ADCpt(EpicsSignalRO, 'MaxShutterOpenDelay_RBV')
    measured_temperature = ADCpt(EpicsSignalRO, 'MeasuredTemperature_RBV')
    min_set_temperature = ADCpt(EpicsSignal, 'MinSetTemperature')
    min_shutter_close_delay = ADCpt(EpicsSignalRO, 'MinShutterCloseDelay_RBV')
    min_shutter_open_delay = ADCpt(EpicsSignalRO, 'MinShutterOpenDelay_RBV')
    num_parallel_pixels = ADCpt(EpicsSignalRO, 'NumParallelPixels_RBV')
    num_ports = ADCpt(EpicsSignalRO, 'NumPorts_RBV')
    num_serial_pixels = ADCpt(EpicsSignalRO, 'NumSerialPixels_RBV')
    num_speed_table_entries = ADCpt(EpicsSignalRO, 'NumSpeedTableEntries_RBV')
    open_delay = ADCpt(SignalWithRBV, 'OpenDelay')
    pcifw_vers = ADCpt(EpicsSignalRO, 'PCIFWVers_RBV')
    pv_cam_vers = ADCpt(EpicsSignalRO, 'PVCamVers_RBV')
    pixel_parallel_dist = ADCpt(EpicsSignalRO, 'PixelParallelDist_RBV')
    pixel_parallel_size = ADCpt(EpicsSignalRO, 'PixelParallelSize_RBV')
    pixel_serial_dist = ADCpt(EpicsSignalRO, 'PixelSerialDist_RBV')
    pixel_serial_size = ADCpt(EpicsSignalRO, 'PixelSerialSize_RBV')
    pixel_time = ADCpt(EpicsSignalRO, 'PixelTime_RBV')
    post_mask = ADCpt(EpicsSignalRO, 'PostMask_RBV')
    post_scan = ADCpt(EpicsSignalRO, 'PostScan_RBV')
    pre_mask = ADCpt(EpicsSignalRO, 'PreMask_RBV')
    pre_scan = ADCpt(EpicsSignalRO, 'PreScan_RBV')
    serial_num = ADCpt(EpicsSignalRO, 'SerialNum_RBV')
    set_temperature = ADCpt(SignalWithRBV, 'SetTemperature')
    slot1_cam = ADCpt(EpicsSignalRO, 'Slot1Cam_RBV')
    slot2_cam = ADCpt(EpicsSignalRO, 'Slot2Cam_RBV')
    slot3_cam = ADCpt(EpicsSignalRO, 'Slot3Cam_RBV')
    speed_table_index = ADCpt(SignalWithRBV, 'SpeedTableIndex')
    trigger_edge = ADCpt(SignalWithRBV, 'TriggerEdge')


class RoperDetectorCam(CamBase):
    _html_docs = ['RoperDoc.html']
    auto_data_type = ADCpt(SignalWithRBV, 'AutoDataType')
    comment1 = ADCpt(SignalWithRBV, 'Comment1')
    comment2 = ADCpt(SignalWithRBV, 'Comment2')
    comment3 = ADCpt(SignalWithRBV, 'Comment3')
    comment4 = ADCpt(SignalWithRBV, 'Comment4')
    comment5 = ADCpt(SignalWithRBV, 'Comment5')
    file_format = ADCpt(SignalWithRBV, 'FileFormat')
    num_acquisitions = ADCpt(SignalWithRBV, 'NumAcquisitions')
    num_acquisitions_counter = ADCpt(EpicsSignalRO, 'NumAcquisitionsCounter_RBV')
    roper_shutter_mode = ADCpt(SignalWithRBV, 'RoperShutterMode')


class URLDetectorCam(CamBase):
    _html_docs = ['URLDoc.html']
    urls = DDC(ad_group(EpicsSignal,
                        (('url_1', 'URL1'),
                         ('url_2', 'URL2'),
                         ('url_3', 'URL3'),
                         ('url_4', 'URL4'),
                         ('url_5', 'URL5'),
                         ('url_6', 'URL6'),
                         ('url_7', 'URL7'),
                         ('url_8', 'URL8'),
                         ('url_9', 'URL9'),
                         ('url_10', 'URL10'))),
               doc='URLs')

    url_select = ADCpt(EpicsSignal, 'URLSelect')
    url_seq = ADCpt(EpicsSignal, 'URLSeq')
    url = ADCpt(EpicsSignalRO, 'URL_RBV')



import logging
import pytest

from ophyd import QuadEM, Kind
from ophyd.sim import (make_fake_device, clear_fake_device)
from ophyd.areadetector.plugins import ImagePlugin, StatsPlugin


logger = logging.getLogger(__name__)


@pytest.fixture(scope='function')
def quadem():
    FakeQuadEM = make_fake_device(QuadEM)
    em = FakeQuadEM('quadem:', name='quadem')
    clear_fake_device(em)

    em.conf.port_name.put('NSLS_EM')

    for k in ['image', 'current1', 'current2', 'current3', 'current4',
              'sum_all']:
        cc = getattr(em, k)

        if isinstance(cc, ImagePlugin):
            cc.plugin_type.sim_put(ImagePlugin._plugin_type)
            cc.nd_array_port.sim_put('NSLS_EM')
        elif isinstance(cc, StatsPlugin):
            cc.plugin_type.sim_put(StatsPlugin._plugin_type)
            cc.nd_array_port.sim_put('NSLS_EM')
        else:
            cc.plugin_type.sim_put('unknown')

        cc.enable.sim_set_enum_strs(['Disabled', 'Enabled'])
        cc.enable.put('Enabled')
        cc.port_name.sim_put(k.upper())

    em.wait_for_connection()

    return em


def test_connected(quadem):
    assert quadem.connected


def test_scan_point(quadem):
    assert quadem._staged.value == 'no'

    quadem.stage()
    assert quadem._staged.value == 'yes'

    quadem.trigger()
    quadem.unstage()
    assert quadem._staged.value == 'no'


def test_reading(quadem):
    assert 'current1.mean_value' in quadem.read_attrs

    vals = quadem.read()
    assert 'quadem_current1_mean_value' in vals
    assert (set(('value', 'timestamp')) ==
            set(vals['quadem_current1_mean_value'].keys()))

    rc = quadem.read_configuration()
    dc = quadem.describe_configuration()

    assert quadem.averaging_time.name in rc
    assert quadem.integration_time.name in rc

    assert rc.keys() == dc.keys()


def test_hints(quadem):

    desc = quadem.describe()
    f_hints = quadem.hints['fields']
    assert len(f_hints) > 0
    for k in f_hints:
        assert k in desc

    def clear_hints(dev):
        for c in dev.component_names:
            c = getattr(dev, c)
            c.kind &= ~(Kind.hinted & ~Kind.normal)
            if hasattr(c, 'component_names'):
                clear_hints(c)

    clear_hints(quadem)

    quadem.current1.mean_value.kind = Kind.hinted

    assert quadem.hints == {'fields': ['quadem_current1_mean_value']}



#!/usr/bin/env python3
from caproto.server import (pvproperty, PVGroup, ioc_arg_parser, run, SubGroup,
                            get_pv_pair_wrapper)

pvproperty_with_rbv = get_pv_pair_wrapper(setpoint_suffix='',
                                          readback_suffix='_RBV')
unknown = int


class MCAROIGroup(PVGroup):
    label = pvproperty(value='label', name='NM')
    count = pvproperty(value=1, name='', read_only=True)
    net_count = pvproperty(name='N', dtype=unknown, read_only=True)
    preset_count = pvproperty(name='P', dtype=unknown)
    is_preset = pvproperty(name='IP', dtype=unknown)
    bkgnd_chans = pvproperty(name='BG', dtype=unknown)
    hi_chan = pvproperty(name='HI', dtype=unknown)
    lo_chan = pvproperty(name='LO', dtype=unknown)


class EpicsMCAGroup(PVGroup):
    stop_signal = pvproperty(name='Stop', dtype=unknown)
    preset_real_time = pvproperty(name='.PRTM', dtype=unknown)
    preset_live_time = pvproperty(name='.PLTM', dtype=unknown)
    elapsed_real_time = pvproperty(name='.ERTM', dtype=unknown, read_only=True)
    elapsed_live_time = pvproperty(name='.ELTM', dtype=unknown, read_only=True)
    spectrum = pvproperty(name='', dtype=float, read_only=True)
    background = pvproperty(name='.BG', dtype=unknown, read_only=True)
    mode = pvproperty(value='List', name='.MODE', dtype=str)

    class RoisGroup(PVGroup):
        roi0 = SubGroup(MCAROIGroup, prefix='.R0')
        roi1 = SubGroup(MCAROIGroup, prefix='.R1')
        roi2 = SubGroup(MCAROIGroup, prefix='.R2')
        roi3 = SubGroup(MCAROIGroup, prefix='.R3')
        roi4 = SubGroup(MCAROIGroup, prefix='.R4')
        roi5 = SubGroup(MCAROIGroup, prefix='.R5')
        roi6 = SubGroup(MCAROIGroup, prefix='.R6')
        roi7 = SubGroup(MCAROIGroup, prefix='.R7')
        roi8 = SubGroup(MCAROIGroup, prefix='.R8')
        roi9 = SubGroup(MCAROIGroup, prefix='.R9')
        roi10 = SubGroup(MCAROIGroup, prefix='.R10')
        roi11 = SubGroup(MCAROIGroup, prefix='.R11')
        roi12 = SubGroup(MCAROIGroup, prefix='.R12')
        roi13 = SubGroup(MCAROIGroup, prefix='.R13')
        roi14 = SubGroup(MCAROIGroup, prefix='.R14')
        roi15 = SubGroup(MCAROIGroup, prefix='.R15')
        roi16 = SubGroup(MCAROIGroup, prefix='.R16')
        roi17 = SubGroup(MCAROIGroup, prefix='.R17')
        roi18 = SubGroup(MCAROIGroup, prefix='.R18')
        roi19 = SubGroup(MCAROIGroup, prefix='.R19')
        roi20 = SubGroup(MCAROIGroup, prefix='.R20')
        roi21 = SubGroup(MCAROIGroup, prefix='.R21')
        roi22 = SubGroup(MCAROIGroup, prefix='.R22')
        roi23 = SubGroup(MCAROIGroup, prefix='.R23')
        roi24 = SubGroup(MCAROIGroup, prefix='.R24')
        roi25 = SubGroup(MCAROIGroup, prefix='.R25')
        roi26 = SubGroup(MCAROIGroup, prefix='.R26')
        roi27 = SubGroup(MCAROIGroup, prefix='.R27')
        roi28 = SubGroup(MCAROIGroup, prefix='.R28')
        roi29 = SubGroup(MCAROIGroup, prefix='.R29')
        roi30 = SubGroup(MCAROIGroup, prefix='.R30')
        roi31 = SubGroup(MCAROIGroup, prefix='.R31')

    rois = SubGroup(RoisGroup, prefix='')

    start = pvproperty(name='Start', dtype=unknown)
    erase = pvproperty(name='Erase', dtype=unknown)
    erase_start = pvproperty(name='EraseStart', dtype=unknown)
    check_acquiring = pvproperty(name='CheckACQG', dtype=unknown)
    client_wait = pvproperty(name='ClientWait', dtype=unknown)
    enable_wait = pvproperty(name='EnableWait', dtype=unknown)
    force_read = pvproperty(name='Read', dtype=unknown)
    set_client_wait = pvproperty(name='SetClientWait', dtype=unknown)
    status = pvproperty(name='Status', dtype=unknown)
    when_acq_stops = pvproperty(name='WhenAcqStops', dtype=unknown)
    why1 = pvproperty(name='Why1', dtype=unknown)
    why2 = pvproperty(name='Why2', dtype=unknown)
    why3 = pvproperty(name='Why3', dtype=unknown)
    why4 = pvproperty(name='Why4', dtype=unknown)


class EpicsDXPGroup(PVGroup):
    preset_mode = pvproperty_with_rbv(value='Live time', name='PresetMode',
                                      dtype=str)
    live_time_output = pvproperty_with_rbv(value='livetimeoutput',
                                           name='LiveTimeOutput', dtype=str)
    elapsed_live_time = pvproperty(name='ElapsedLiveTime', dtype=unknown)
    elapsed_real_time = pvproperty(name='ElapsedRealTime', dtype=unknown)
    elapsed_trigger_live_time = pvproperty(
        name='ElapsedTriggerLiveTime', dtype=unknown)
    trigger_peaking_time = pvproperty_with_rbv(
        name='TriggerPeakingTime', dtype=unknown)
    trigger_threshold = pvproperty_with_rbv(
        name='TriggerThreshold', dtype=unknown)
    trigger_gap_time = pvproperty_with_rbv(
        name='TriggerGapTime', dtype=unknown)
    trigger_output = pvproperty_with_rbv(value='trigger_output',
                                         name='TriggerOutput', dtype=str)
    max_width = pvproperty_with_rbv(name='MaxWidth', dtype=unknown)
    peaking_time = pvproperty_with_rbv(name='PeakingTime', dtype=unknown)
    energy_threshold = pvproperty_with_rbv(
        name='EnergyThreshold', dtype=unknown)
    gap_time = pvproperty_with_rbv(name='GapTime', dtype=unknown)
    baseline_cut_percent = pvproperty_with_rbv(
        name='BaselineCutPercent', dtype=unknown)
    baseline_cut_enable = pvproperty_with_rbv(
        name='BaselineCutEnable', dtype=unknown)
    baseline_filter_length = pvproperty_with_rbv(
        name='BaselineFilterLength', dtype=unknown)
    baseline_threshold = pvproperty_with_rbv(
        name='BaselineThreshold', dtype=unknown)
    baseline_energy_array = pvproperty(
        name='BaselineEnergyArray', dtype=unknown)
    baseline_histogram = pvproperty(name='BaselineHistogram', dtype=unknown)
    preamp_gain = pvproperty_with_rbv(name='PreampGain', dtype=unknown)
    detector_polarity = pvproperty_with_rbv(
        name='DetectorPolarity', dtype=unknown)
    reset_delay = pvproperty_with_rbv(name='ResetDelay', dtype=unknown)
    decay_time = pvproperty_with_rbv(name='DecayTime', dtype=unknown)
    max_energy = pvproperty_with_rbv(name='MaxEnergy', dtype=unknown)
    adc_percent_rule = pvproperty_with_rbv(
        name='ADCPercentRule', dtype=unknown)
    triggers = pvproperty(name='Triggers', dtype=unknown, read_only=True)
    events = pvproperty(name='Events', dtype=unknown, read_only=True)
    overflows = pvproperty(name='Overflows', dtype=unknown, read_only=True)
    underflows = pvproperty(name='Underflows', dtype=unknown, read_only=True)
    input_count_rate = pvproperty(
        name='InputCountRate', dtype=unknown, read_only=True)
    output_count_rate = pvproperty(
        name='OutputCountRate', dtype=unknown, read_only=True)
    mca_bin_width = pvproperty(
        name='MCABinWidth_RBV', dtype=unknown, read_only=True)
    calibration_energy = pvproperty(
        name='CalibrationEnergy_RBV', dtype=unknown, read_only=True)
    current_pixel = pvproperty(name='CurrentPixel', dtype=unknown)
    dynamic_range = pvproperty(
        name='DynamicRange_RBV', dtype=unknown, read_only=True)
    preset_events = pvproperty_with_rbv(name='PresetEvents', dtype=unknown)
    preset_triggers = pvproperty_with_rbv(name='PresetTriggers', dtype=unknown)
    trace_data = pvproperty(name='TraceData', dtype=unknown)
    trace_mode = pvproperty_with_rbv(value='Mode', name='TraceMode', dtype=str)
    trace_time_array = pvproperty(name='TraceTimeArray', dtype=unknown)
    trace_time = pvproperty_with_rbv(name='TraceTime', dtype=unknown)


class McaDxpIOC(PVGroup):
    mca = SubGroup(EpicsMCAGroup, prefix='mca')
    dxp = SubGroup(EpicsDXPGroup, prefix='dxp:')


if __name__ == '__main__':
    ioc_options, run_options = ioc_arg_parser(
        default_prefix='test_mca:',
        desc="ophyd.tests.test_mca test IOC")
    ioc = McaDxpIOC(**ioc_options)
    run(ioc.pvdb, **run_options)



import logging
import time
import uuid

import pytest

from types import SimpleNamespace

from ophyd import (set_cl, EpicsMotor, Signal, EpicsSignal, EpicsSignalRO,
                   Component as Cpt)
from ophyd.utils.epics_pvs import (AlarmSeverity, AlarmStatus)

logger = logging.getLogger(__name__)


@pytest.fixture()
def hw(tmpdir):
    from ophyd.sim import hw
    return hw(str(tmpdir))


@pytest.fixture(params=['caproto', 'pyepics'], autouse=True,
                scope='session')
def cl_selector(request):
    cl_name = request.param
    if cl_name == 'caproto':
        pytest.importorskip('caproto')
        logging.getLogger('caproto.bcast').setLevel('INFO')
    elif cl_name == 'pyepics':
        pytest.importorskip('epics')
    set_cl(cl_name)
    yield
    set_cl()


class CustomAlarmEpicsSignalRO(EpicsSignalRO):
    alarm_status = AlarmStatus.NO_ALARM
    alarm_severity = AlarmSeverity.NO_ALARM


class TestEpicsMotor(EpicsMotor):
    user_readback = Cpt(CustomAlarmEpicsSignalRO, '.RBV', kind='hinted')
    high_limit_switch = Cpt(Signal, value=0, kind='omitted')
    low_limit_switch = Cpt(Signal, value=0, kind='omitted')
    direction_of_travel = Cpt(Signal, value=0, kind='omitted')
    high_limit_value = Cpt(EpicsSignal, '.HLM', kind='config')
    low_limit_value = Cpt(EpicsSignal, '.LLM', kind='config')

    @user_readback.sub_value
    def _pos_changed(self, timestamp=None, value=None, **kwargs):
        '''Callback from EPICS, indicating a change in position'''
        super()._pos_changed(timestamp=timestamp, value=value, **kwargs)


@pytest.fixture(scope='function')
def motor(request, cleanup):
    sim_pv = 'XF:31IDA-OP{Tbl-Ax:X1}Mtr'

    motor = TestEpicsMotor(sim_pv, name='epicsmotor', settle_time=0.1,
                           timeout=10.0)
    cleanup.add(motor)

    print('Created EpicsMotor:', motor)
    motor.wait_for_connection()
    motor.low_limit_value.put(-100, wait=True)
    motor.high_limit_value.put(100, wait=True)
    # set the motor to 0
    motor.set(0)
    while motor.motor_done_move.get() != 1:
        print('Waiting for {} to stop moving...'.format(motor))
        time.sleep(0.5)

    return motor


@pytest.fixture(scope='module')
def ad_prefix():
    'AreaDetector prefix'
    prefixes = ['13SIM1:', 'XF:31IDA-BI{Cam:Tbl}']

    for prefix in prefixes:
        test_pv = prefix + 'TIFF1:PluginType_RBV'
        try:
            sig = EpicsSignalRO(test_pv)
            sig.wait_for_connection(timeout=2)
        except TimeoutError:
            ...
        else:
            print('areaDetector detected with prefix:', prefix)
            return prefix
        finally:
            sig.destroy()
    raise pytest.skip('No areaDetector IOC running')


@pytest.fixture(scope='function')
def prefix():
    'Random PV prefix for a server'
    return str(uuid.uuid4())[:8] + ':'


@pytest.fixture(scope='function')
def fake_motor_ioc(prefix, request):
    name = 'Fake motor IOC'
    pvs = dict(setpoint=f'{prefix}setpoint',
               readback=f'{prefix}readback',
               moving=f'{prefix}moving',
               actuate=f'{prefix}actuate',
               stop=f'{prefix}stop',
               step_size=f'{prefix}step_size',
               )

    pytest.importorskip('caproto.tests.conftest')
    from caproto.tests.conftest import run_example_ioc
    process = run_example_ioc('ophyd.tests.fake_motor_ioc',
                              request=request,
                              pv_to_check=pvs['setpoint'],
                              args=('--prefix', prefix, '--list-pvs', '-v'))
    return SimpleNamespace(process=process, prefix=prefix, name=name, pvs=pvs,
                           type='caproto')


@pytest.fixture(scope='function')
def signal_test_ioc(prefix, request):
    name = 'test_signal IOC'
    pvs = dict(read_only=f'{prefix}read_only',
               read_write=f'{prefix}read_write',
               pair_set=f'{prefix}pair_set',
               pair_rbv=f'{prefix}pair_rbv',
               waveform=f'{prefix}waveform',
               bool_enum=f'{prefix}bool_enum',
               alarm_status=f'{prefix}alarm_status',
               set_severity=f'{prefix}set_severity',
               )

    pytest.importorskip('caproto.tests.conftest')
    from caproto.tests.conftest import run_example_ioc
    process = run_example_ioc('ophyd.tests.signal_ioc',
                              request=request,
                              pv_to_check=pvs['read_only'],
                              args=('--prefix', prefix, '--list-pvs', '-v'))
    return SimpleNamespace(process=process, prefix=prefix, name=name, pvs=pvs,
                           type='caproto')


@pytest.fixture(scope='function')
def cleanup(request):
    'Destroy all items added to the list during the finalizer'
    items = []

    class Cleaner:
        def add(self, item):
            items.append(item)

    def clean():
        for item in items:
            print('Destroying', item.name)
            item.destroy()
        items.clear()

    request.addfinalizer(clean)
    return Cleaner()



import time
import functools
import logging
import queue
import threading


class _CallbackThread(threading.Thread):
    'A queue-based callback dispatcher thread'

    def __init__(self, name, *, dispatcher, logger, context,
                 stop_event, timeout, callback_queue=None, daemon=True):
        super().__init__(name=name, daemon=daemon)
        self.context = context
        self.current_callback = None
        self.dispatcher = dispatcher
        self.logger = logger
        self.stop_event = stop_event
        self.timeout = timeout

        if callback_queue is None:
            callback_queue = queue.Queue()

        self.queue = callback_queue

    def __repr__(self):
        return '<{} qsize={}>'.format(self.__class__.__name__,
                                      self.queue.qsize())

    def run(self):
        '''The dispatcher itself'''
        self.logger.debug('Callback thread %s started', self.name)
        self.attach_context()

        while not self.stop_event.is_set():
            try:
                callback, args, kwargs = self.queue.get(True, self.timeout)
            except queue.Empty:
                ...
            else:
                try:
                    self.current_callback = (
                        getattr(callback, '__name__', '(unnamed)'),
                        kwargs.get('pvname'))
                    callback(*args, **kwargs)
                except Exception:
                    self.logger.exception(
                        'Exception occurred during callback %r (pvname=%r)',
                        callback, kwargs.get('pvname')
                    )

        self.detach_context()

    def attach_context(self):
        self.logger.debug('Callback thread %s attaching to context %s',
                          self.name, self.context)

    def detach_context(self):
        self.context = None


class DispatcherThreadContext:
    '''
    A thread context associated with a single Dispatcher event type

    Parameters
    ----------
    dispatcher : Dispatcher
    event_type : str

    Attributes
    ----------
    dispatcher : Dispatcher
    event_type : str
    event_thread : _CallbackThread
    '''

    def __init__(self, dispatcher, event_type):
        self.dispatcher = dispatcher
        self.event_type = event_type
        self.event_thread = None

    def run(self, func, *args, **kwargs):
        '''
        If in the correct threading context, run func(*args, **kwargs) directly,
        otherwise schedule it to be run in that thread.
        '''
        if self.event_thread is None:
            self.event_thread = self.dispatcher._threads[self.event_type]

        current_thread = threading.currentThread()
        if current_thread is self.event_thread:
            func(*args, **kwargs)
        else:
            self.event_thread.queue.put((func, args, kwargs))

    __call__ = run


debug_monitor_log = logging.getLogger('ophyd.event_dispatcher')


class EventDispatcher:
    def __init__(self, *, context, logger, timeout=0.1,
                 thread_class=_CallbackThread,
                 utility_threads=4):
        self._threads = {}
        self._thread_contexts = {}
        self._thread_class = thread_class
        self._timeout = timeout

        # The dispatcher thread will stop if this event is set
        self._stop_event = threading.Event()
        self.context = context
        self.logger = logger
        self.debug_monitor_interval = 1
        self._utility_threads = [f'util{i}' for i in range(utility_threads)]
        self._utility_queue = queue.Queue()

        self._start_thread(name='metadata')
        self._start_thread(name='monitor')
        self._start_thread(name='get_put')

        for name in self._utility_threads:
            self._start_thread(name=name, callback_queue=self._utility_queue)

        self._debug_monitor_thread = threading.Thread(
            target=self._debug_monitor,
            name='debug_monitor',
            daemon=True)
        self._debug_monitor_thread.start()

    def _debug_monitor(self):
        while not self._stop_event.is_set():
            queue_sizes = [(name, thread.queue.qsize(), thread.current_callback)
                           for name, thread in sorted(self._threads.items())
                           ]
            status = [
                '{name}={qsize} ({cb})'.format(name=name, qsize=qsize, cb=cb)
                for name, qsize, cb in queue_sizes
                if qsize
            ]
            if status:
                debug_monitor_log.debug(' / '.join(status))
            else:
                debug_monitor_log.debug('All EventDispatch queues are empty.')
            time.sleep(self.debug_monitor_interval)

    def __repr__(self):
        threads = [repr(thread) for thread in self._threads.values()]
        return '<{} threads={}>'.format(self.__class__.__name__, threads)

    def is_alive(self):
        return any(thread.is_alive() for thread in self.threads.values()
                   if thread is not None)

    @property
    def stop_event(self):
        return self._stop_event

    @property
    def timeout(self):
        return self._timeout

    @property
    def threads(self):
        return dict(self._threads)

    def stop(self):
        '''Stop the dispatcher threads and re-enable normal callbacks'''
        self._stop_event.set()
        for attr, thread in list(self._threads.items()):
            if thread is not None:
                thread.join()

        self._threads.clear()
        self._debug_monitor_thread.join()

    def schedule_utility_task(self, callback, *args, **kwargs):
        'Schedule `callback` with the given args and kwargs in a util thread'
        self._utility_queue.put((callback, args, kwargs))

    def get_thread_context(self, name):
        'Get the DispatcherThreadContext for the given thread name'
        return self._thread_contexts[name]

    def _start_thread(self, name, *, callback_queue=None):
        'Start dispatcher thread by name'
        self._threads[name] = self._thread_class(name=name, dispatcher=self,
                                                 stop_event=self._stop_event,
                                                 timeout=self.timeout,
                                                 context=self.context,
                                                 logger=self.logger,
                                                 daemon=True,
                                                 callback_queue=callback_queue)
        self._thread_contexts[name] = DispatcherThreadContext(self, name)
        self._threads[name].start()


def wrap_callback(dispatcher, event_type, callback):
    'Wrap a callback for usage with the dispatcher'
    if callback is None or getattr(callback, '_wrapped_callback', False):
        return callback

    assert event_type in dispatcher._threads
    callback_queue = dispatcher._threads[event_type].queue

    @functools.wraps(callback)
    def wrapped(*args, **kwargs):
        callback_queue.put((callback, args, kwargs))

    wrapped._wrapped_callback = True
    return wrapped




import logging

from collections import OrderedDict

from .signal import (Signal, EpicsSignal, EpicsSignalRO)
from .device import (Device, Component as Cpt, DynamicDeviceComponent as DDC,
                     BlueskyInterface, Kind)
from .areadetector import EpicsSignalWithRBV as SignalWithRBV


logger = logging.getLogger(__name__)


class ROI(Device):

    # 'name' is not an allowed attribute
    label = Cpt(EpicsSignal, 'NM', lazy=True)
    count = Cpt(EpicsSignalRO, '', lazy=True)
    net_count = Cpt(EpicsSignalRO, 'N', lazy=True)
    preset_count = Cpt(EpicsSignal, 'P', lazy=True)
    is_preset = Cpt(EpicsSignal, 'IP', lazy=True)
    bkgnd_chans = Cpt(EpicsSignal, 'BG', lazy=True)
    hi_chan = Cpt(EpicsSignal, 'HI', lazy=True)
    lo_chan = Cpt(EpicsSignal, 'LO', lazy=True)

    def __init__(self, prefix, *, read_attrs=None, configuration_attrs=None,
                 name=None, parent=None, **kwargs):

        super().__init__(prefix, read_attrs=read_attrs,
                         configuration_attrs=configuration_attrs,
                         name=name, parent=parent, **kwargs)


def add_rois(range_, **kwargs):
    '''Add one or more ROIs to an MCA instance

       Parameters
       ----------
       range_ : sequence of ints
           Must be be in the set [0,31]

       By default, an EpicsMCA is initialized with all 32 rois.
       These provide the following Components as EpicsSignals (N=[0,31]):
       EpicsMCA.rois.roiN.(label,count,net_count,preset_cnt, is_preset,
       bkgnd_chans, hi_chan, lo_chan)
       '''
    defn = OrderedDict()

    for roi in range_:
        if not (0 <= roi < 32):
            raise ValueError('roi must be in the set [0,31]')

        attr = 'roi{}'.format(roi)
        defn[attr] = (ROI, '.R{}'.format(roi), kwargs)

    return defn


class EpicsMCARecord(Device):
    '''SynApps MCA Record interface'''
    stop_signal = Cpt(EpicsSignal, '.STOP', kind='omitted')
    preset_real_time = Cpt(EpicsSignal, '.PRTM', kind=Kind.config | Kind.normal)
    preset_live_time = Cpt(EpicsSignal, '.PLTM', kind='omitted')
    elapsed_real_time = Cpt(EpicsSignalRO, '.ERTM')
    elapsed_live_time = Cpt(EpicsSignalRO, '.ELTM', kind='omitted')

    spectrum = Cpt(EpicsSignalRO, '.VAL')
    background = Cpt(EpicsSignalRO, '.BG', kind='omitted')
    mode = Cpt(EpicsSignal, '.MODE', string=True, kind='omitted')

    rois = DDC(add_rois(range(0, 32), kind='omitted'), kind='omitted')

    def __init__(self, *args, **kwargs):

        super().__init__(*args, **kwargs)

        # could arguably be made a configuration_attr instead...
        self.stage_sigs['mode'] = 'PHA'

    def stop(self, *, success=False):
        self.stop_signal.put(1)


class EpicsMCA(EpicsMCARecord):
    '''mca records with extras from mca.db'''
    start = Cpt(EpicsSignal, 'Start', kind='omitted')
    stop_signal = Cpt(EpicsSignal, 'Stop', kind='omitted')
    erase = Cpt(EpicsSignal, 'Erase', kind='omitted')
    erase_start = Cpt(EpicsSignal, 'EraseStart', trigger_value=1, kind='omitted')

    check_acquiring = Cpt(EpicsSignal, 'CheckACQG', kind='omitted')
    client_wait = Cpt(EpicsSignal, 'ClientWait', kind='omitted')
    enable_wait = Cpt(EpicsSignal, 'EnableWait', kind='omitted')
    force_read = Cpt(EpicsSignal, 'Read', kind='omitted')
    set_client_wait = Cpt(EpicsSignal, 'SetClientWait', kind='omitted')
    status = Cpt(EpicsSignal, 'Status', kind='omitted')
    when_acq_stops = Cpt(EpicsSignal, 'WhenAcqStops', kind='omitted')
    why1 = Cpt(EpicsSignal, 'Why1', kind='omitted')
    why2 = Cpt(EpicsSignal, 'Why2', kind='omitted')
    why3 = Cpt(EpicsSignal, 'Why3', kind='omitted')
    why4 = Cpt(EpicsSignal, 'Why4', kind='omitted')


class EpicsMCAReadNotify(EpicsMCARecord):
    '''mca record with extras from mcaReadNotify.db'''
    start = Cpt(EpicsSignal, 'Start', kind='omitted')
    stop_signal = Cpt(EpicsSignal, 'Stop', kind='omitted')
    erase = Cpt(EpicsSignal, 'Erase', kind='omitted')
    erase_start = Cpt(EpicsSignal, 'EraseStart', trigger_value=1, kind='omitted')

    check_acquiring = Cpt(EpicsSignal, 'CheckACQG', kind='omitted')
    client_wait = Cpt(EpicsSignal, 'ClientWait', kind='omitted')
    enable_wait = Cpt(EpicsSignal, 'EnableWait', kind='omitted')
    force_read = Cpt(EpicsSignal, 'Read', kind='omitted')
    set_client_wait = Cpt(EpicsSignal, 'SetClientWait', kind='omitted')
    status = Cpt(EpicsSignal, 'Status', kind='omitted')


class EpicsMCACallback(Device):
    '''Callback-related signals for MCA devices'''
    read_callback = Cpt(EpicsSignal, 'ReadCallback')
    read_data_once = Cpt(EpicsSignal, 'ReadDataOnce')
    read_status_once = Cpt(EpicsSignal, 'ReadStatusOnce')
    collect_data = Cpt(EpicsSignal, 'CollectData')


class EpicsDXP(Device):
    '''All high-level DXP parameters for each channel'''
    preset_mode = Cpt(EpicsSignal, 'PresetMode', string=True)

    live_time_output = Cpt(SignalWithRBV, 'LiveTimeOutput', string=True)
    elapsed_live_time = Cpt(EpicsSignal, 'ElapsedLiveTime')
    elapsed_real_time = Cpt(EpicsSignal, 'ElapsedRealTime')
    elapsed_trigger_live_time = Cpt(EpicsSignal, 'ElapsedTriggerLiveTime')

    # Trigger Filter PVs
    trigger_peaking_time = Cpt(SignalWithRBV, 'TriggerPeakingTime')
    trigger_threshold = Cpt(SignalWithRBV, 'TriggerThreshold')
    trigger_gap_time = Cpt(SignalWithRBV, 'TriggerGapTime')
    trigger_output = Cpt(SignalWithRBV, 'TriggerOutput', string=True)
    max_width = Cpt(SignalWithRBV, 'MaxWidth')

    # Energy Filter PVs
    peaking_time = Cpt(SignalWithRBV, 'PeakingTime')
    energy_threshold = Cpt(SignalWithRBV, 'EnergyThreshold')
    gap_time = Cpt(SignalWithRBV, 'GapTime')

    # Baseline PVs
    baseline_cut_percent = Cpt(SignalWithRBV, 'BaselineCutPercent')
    baseline_cut_enable = Cpt(SignalWithRBV, 'BaselineCutEnable')
    baseline_filter_length = Cpt(SignalWithRBV, 'BaselineFilterLength')
    baseline_threshold = Cpt(SignalWithRBV, 'BaselineThreshold')
    baseline_energy_array = Cpt(EpicsSignal, 'BaselineEnergyArray')
    baseline_histogram = Cpt(EpicsSignal, 'BaselineHistogram')
    baseline_threshold = Cpt(SignalWithRBV, 'BaselineThreshold')

    # Misc PVs
    preamp_gain = Cpt(SignalWithRBV, 'PreampGain')
    detector_polarity = Cpt(SignalWithRBV, 'DetectorPolarity')
    reset_delay = Cpt(SignalWithRBV, 'ResetDelay')
    decay_time = Cpt(SignalWithRBV, 'DecayTime')
    max_energy = Cpt(SignalWithRBV, 'MaxEnergy')
    adc_percent_rule = Cpt(SignalWithRBV, 'ADCPercentRule')
    max_width = Cpt(SignalWithRBV, 'MaxWidth')

    # read-only diagnostics
    triggers = Cpt(EpicsSignalRO, 'Triggers', lazy=True)
    events = Cpt(EpicsSignalRO, 'Events', lazy=True)
    overflows = Cpt(EpicsSignalRO, 'Overflows', lazy=True)
    underflows = Cpt(EpicsSignalRO, 'Underflows', lazy=True)
    input_count_rate = Cpt(EpicsSignalRO, 'InputCountRate', lazy=True)
    output_count_rate = Cpt(EpicsSignalRO, 'OutputCountRate', lazy=True)

    mca_bin_width = Cpt(EpicsSignalRO, 'MCABinWidth_RBV')
    calibration_energy = Cpt(EpicsSignalRO, 'CalibrationEnergy_RBV')
    current_pixel = Cpt(EpicsSignal, 'CurrentPixel')
    dynamic_range = Cpt(EpicsSignalRO, 'DynamicRange_RBV')

    # Preset options
    preset_events = Cpt(SignalWithRBV, 'PresetEvents')
    preset_mode = Cpt(SignalWithRBV, 'PresetMode', string=True)
    preset_triggers = Cpt(SignalWithRBV, 'PresetTriggers')

    # Trace options
    trace_data = Cpt(EpicsSignal, 'TraceData')
    trace_mode = Cpt(SignalWithRBV, 'TraceMode', string=True)
    trace_time_array = Cpt(EpicsSignal, 'TraceTimeArray')
    trace_time = Cpt(SignalWithRBV, 'TraceTime')


class EpicsDXPLowLevelParameter(Device):
    param_name = Cpt(EpicsSignal, 'Name')
    value = Cpt(SignalWithRBV, 'Val')


class EpicsDXPLowLevel(Device):
    num_low_level_params = Cpt(EpicsSignal, 'NumLLParams')
    read_low_level_params = Cpt(EpicsSignal, 'ReadLLParams')

    parameter_prefix = 'LL{}'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._parameter_cache = {}

    def get_low_level_parameter(self, index):
        '''Get a DXP low level parameter

        Parameters
        ----------
        index : int
            In the range of [0, 229]

        Returns
        -------
        param : EpicsDXPLowLevelParameter
        '''
        try:
            return self._parameter_cache[index]
        except KeyError:
            pass

        prefix = '{}{}'.format(self.prefix, self.parameter_prefix)
        name = '{}_param{}'.format(self.name, index)
        param = EpicsDXPLowLevelParameter(prefix, name=name)
        self._parameter_cache[index] = param
        return param


class EpicsDXPMapping(Device):
    apply = Cpt(EpicsSignal, 'Apply')
    auto_apply = Cpt(SignalWithRBV, 'AutoApply')
    auto_pixels_per_buffer = Cpt(SignalWithRBV, 'AutoPixelsPerBuffer')
    buffer_size = Cpt(EpicsSignalRO, 'BufferSize_RBV')
    collect_mode = Cpt(SignalWithRBV, 'CollectMode')
    ignore_gate = Cpt(SignalWithRBV, 'IgnoreGate')
    input_logic_polarity = Cpt(SignalWithRBV, 'InputLogicPolarity')
    list_mode = Cpt(SignalWithRBV, 'ListMode')
    mbytes_read = Cpt(EpicsSignalRO, 'MBytesRead_RBV')
    next_pixel = Cpt(EpicsSignal, 'NextPixel')
    pixel_advance_mode = Cpt(SignalWithRBV, 'PixelAdvanceMode')
    pixels_per_buffer = Cpt(SignalWithRBV, 'PixelsPerBuffer')
    pixels_per_run = Cpt(SignalWithRBV, 'PixelsPerRun')
    read_rate = Cpt(EpicsSignalRO, 'ReadRate_RBV')
    sync_count = Cpt(SignalWithRBV, 'SyncCount')


class EpicsDXPBaseSystem(Device):
    channel_advance = Cpt(EpicsSignal, 'ChannelAdvance')
    client_wait = Cpt(EpicsSignal, 'ClientWait')
    dwell = Cpt(EpicsSignal, 'Dwell')
    max_scas = Cpt(EpicsSignal, 'MaxSCAs')
    num_scas = Cpt(SignalWithRBV, 'NumSCAs')
    poll_time = Cpt(SignalWithRBV, 'PollTime')
    prescale = Cpt(EpicsSignal, 'Prescale')
    save_system = Cpt(SignalWithRBV, 'SaveSystem')
    save_system_file = Cpt(EpicsSignal, 'SaveSystemFile')
    set_client_wait = Cpt(EpicsSignal, 'SetClientWait')


class EpicsDXPMultiElementSystem(EpicsDXPBaseSystem):
    # Preset info
    preset_events = Cpt(EpicsSignal, 'PresetEvents')
    preset_live_time = Cpt(EpicsSignal, 'PresetLive')
    preset_real_time = Cpt(EpicsSignal, 'PresetReal')
    preset_mode = Cpt(EpicsSignal, 'PresetMode', string=True)
    preset_triggers = Cpt(EpicsSignal, 'PresetTriggers')

    # Acquisition
    erase_all = Cpt(EpicsSignal, 'EraseAll')
    erase_start = Cpt(EpicsSignal, 'EraseStart', trigger_value=1)
    start_all = Cpt(EpicsSignal, 'StartAll')
    stop_all = Cpt(EpicsSignal, 'StopAll')

    # Status
    set_acquire_busy = Cpt(EpicsSignal, 'SetAcquireBusy')
    acquire_busy = Cpt(EpicsSignal, 'AcquireBusy')
    status_all = Cpt(EpicsSignal, 'StatusAll')
    status_all_once = Cpt(EpicsSignal, 'StatusAllOnce')
    acquiring = Cpt(EpicsSignal, 'Acquiring')

    # Reading
    read_baseline_histograms = Cpt(EpicsSignal, 'ReadBaselineHistograms')
    read_all = Cpt(EpicsSignal, 'ReadAll')
    read_all_once = Cpt(EpicsSignal, 'ReadAllOnce')

    # As a debugging note, if snl_connected is not '1', your IOC is
    # misconfigured:
    snl_connected = Cpt(EpicsSignal, 'SNL_Connected')

    # Copying to individual elements
    copy_adcp_ercent_rule = Cpt(EpicsSignal, 'CopyADCPercentRule')
    copy_baseline_cut_enable = Cpt(EpicsSignal, 'CopyBaselineCutEnable')
    copy_baseline_cut_percent = Cpt(EpicsSignal, 'CopyBaselineCutPercent')
    copy_baseline_filter_length = Cpt(EpicsSignal, 'CopyBaselineFilterLength')
    copy_baseline_threshold = Cpt(EpicsSignal, 'CopyBaselineThreshold')
    copy_decay_time = Cpt(EpicsSignal, 'CopyDecayTime')
    copy_detector_polarity = Cpt(EpicsSignal, 'CopyDetectorPolarity')
    copy_energy_threshold = Cpt(EpicsSignal, 'CopyEnergyThreshold')
    copy_gap_time = Cpt(EpicsSignal, 'CopyGapTime')
    copy_max_energy = Cpt(EpicsSignal, 'CopyMaxEnergy')
    copy_max_width = Cpt(EpicsSignal, 'CopyMaxWidth')
    copy_peaking_time = Cpt(EpicsSignal, 'CopyPeakingTime')
    copy_preamp_gain = Cpt(EpicsSignal, 'CopyPreampGain')
    copy_roic_hannel = Cpt(EpicsSignal, 'CopyROIChannel')
    copy_roie_nergy = Cpt(EpicsSignal, 'CopyROIEnergy')
    copy_roi_sca = Cpt(EpicsSignal, 'CopyROI_SCA')
    copy_reset_delay = Cpt(EpicsSignal, 'CopyResetDelay')
    copy_trigger_gap_time = Cpt(EpicsSignal, 'CopyTriggerGapTime')
    copy_trigger_peaking_time = Cpt(EpicsSignal, 'CopyTriggerPeakingTime')
    copy_trigger_threshold = Cpt(EpicsSignal, 'CopyTriggerThreshold')

    # do_* executes the process:
    do_read_all = Cpt(EpicsSignal, 'DoReadAll')
    do_read_baseline_histograms = Cpt(EpicsSignal, 'DoReadBaselineHistograms')
    do_read_traces = Cpt(EpicsSignal, 'DoReadTraces')
    do_status_all = Cpt(EpicsSignal, 'DoStatusAll')

    # Time
    dead_time = Cpt(EpicsSignal, 'DeadTime')
    elapsed_live = Cpt(EpicsSignal, 'ElapsedLive')
    elapsed_real = Cpt(EpicsSignal, 'ElapsedReal')
    idead_time = Cpt(EpicsSignal, 'IDeadTime')

    # low-level
    read_low_level_params = Cpt(EpicsSignal, 'ReadLLParams')

    # Traces
    read_traces = Cpt(EpicsSignal, 'ReadTraces')
    trace_modes = Cpt(EpicsSignal, 'TraceModes', string=True)
    trace_times = Cpt(EpicsSignal, 'TraceTimes')


class SaturnMCA(EpicsMCA, EpicsMCACallback):
    pass


class SaturnDXP(EpicsDXP, EpicsDXPLowLevel):
    pass


class Saturn(EpicsDXPBaseSystem):
    '''DXP Saturn with 1 channel example'''
    dxp = Cpt(SaturnDXP, 'dxp1:')
    mca = Cpt(SaturnMCA, 'mca1')


class MercuryDXP(EpicsDXP, EpicsDXPLowLevel):
    pass


class Mercury1(EpicsDXPMultiElementSystem):
    '''DXP Mercury with 1 channel example'''
    dxp = Cpt(MercuryDXP, 'dxp1:')
    mca = Cpt(EpicsMCARecord, 'mca1')


class SoftDXPTrigger(BlueskyInterface):
    '''Simple soft trigger for DXP devices

    Parameters
    ----------
    count_signal : str, optional
        Signal to set acquisition time (default: 'preset_real_time')
    preset_mode : str, optional
        Default preset mode for the stage signals (default: 'Real time')
    mode_signal : str, optional
        Preset mode signal attribute (default 'preset_mode')
    stop_signal : str, optional
        Stop signal attribute (default 'stop_all')
    '''

    count_time = Cpt(Signal, value=None, doc='bluesky count time')

    def __init__(self, *args, count_signal='preset_real_time',
                 stop_signal='stop_all', mode_signal='preset_mode',
                 preset_mode='Real time',
                 **kwargs):
        super().__init__(*args, **kwargs)
        self._status = None
        self._count_signal = getattr(self, count_signal)

        stop_signal = getattr(self, stop_signal)
        self.stage_sigs[stop_signal] = 1

        mode_signal = getattr(self, mode_signal)
        self.stage_sigs[mode_signal] = preset_mode

    def stage(self):
        if self.count_time.get() is None:
            # remove count_time from the stage signals if count_time unset
            try:
                del self.stage_sigs[self._count_signal]
            except KeyError:
                pass
        else:
            self.stage_sigs[self._count_signal] = self.count_time.get()

        super().stage()



from collections import OrderedDict

from . import (EpicsSignalRO, EpicsSignal, Component as Cpt,
               DynamicDeviceComponent as DDCpt, Signal,
               Kind, kind_context)
from .areadetector import (EpicsSignalWithRBV,
                           ImagePlugin, StatsPlugin, DetectorBase,
                           SingleTrigger, ADBase)
from .status import DeviceStatus


def _current_fields(attr_base, field_base, range_, **kwargs):
    defn = OrderedDict()
    for i in range_:
        attr = '{attr}{i}'.format(attr=attr_base, i=i)
        suffix = '{field}{i}'.format(field=field_base, i=i)
        defn[attr] = (EpicsSignal, suffix, kwargs)

    return defn


class QuadEMPort(ADBase):
    port_name = Cpt(Signal, value='')

    def __init__(self, port_name, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.port_name.put(port_name)


class QuadEM(SingleTrigger, DetectorBase):
    # These settings intentionally shadow the settings inherited from
    # DetectorBase.
    _default_read_attrs = None
    _default_configuration_attrs = None

    _status_type = DeviceStatus  # overrriding the default in SingleTrigger

    # This is needed because ophyd verifies that it can see all
    # of the nodes in the asyn pipeline, however these IOCs do not
    # expose their port name via a PV, but nevertheless server as the
    # root node for the plugins.
    # Leaving this port_name here for compatibility
    integration_time = Cpt(EpicsSignalWithRBV, 'IntegrationTime',
                           kind='config')
    averaging_time = Cpt(EpicsSignalWithRBV, 'AveragingTime', kind='config')
    with kind_context('omitted') as OCpt:
        conf = Cpt(QuadEMPort, port_name='NSLS_EM')
        model = OCpt(EpicsSignalRO, 'Model')
        firmware = OCpt(EpicsSignalRO, 'Firmware')

        acquire_mode = OCpt(EpicsSignalWithRBV, 'AcquireMode')
        acquire = OCpt(EpicsSignal, 'Acquire')

        read_format = OCpt(EpicsSignalWithRBV, 'ReadFormat')
        em_range = OCpt(EpicsSignalWithRBV, 'Range')
        ping_pong = OCpt(EpicsSignalWithRBV, 'PingPong')

        num_channels = OCpt(EpicsSignalWithRBV, 'NumChannels')
        geometry = OCpt(EpicsSignalWithRBV, 'Geometry')
        resolution = OCpt(EpicsSignalWithRBV, 'Resolution')

        bias_state = OCpt(EpicsSignalWithRBV, 'BiasState')
        bias_interlock = OCpt(EpicsSignalWithRBV, 'BiasInterlock')
        bias_voltage = OCpt(EpicsSignalWithRBV, 'BiasVoltage')
        hvs_readback = OCpt(EpicsSignalRO, 'HVSReadback')
        hvv_readback = OCpt(EpicsSignalRO, 'HVVReadback')
        hvi_readback = OCpt(EpicsSignalRO, 'HVIReadback')

        values_per_read = OCpt(EpicsSignalWithRBV, 'ValuesPerRead')
        sample_time = OCpt(EpicsSignalRO, 'SampleTime_RBV')  # yay consistency
        num_average = OCpt(EpicsSignalRO, 'NumAverage_RBV')
        num_averaged = OCpt(EpicsSignalRO, 'NumAveraged_RBV')
        num_acquire = OCpt(EpicsSignalWithRBV, 'NumAcquire')
        num_acquired = OCpt(EpicsSignalRO, 'NumAcquired')
        read_data = OCpt(EpicsSignalRO, 'ReadData')
        ring_overflows = OCpt(EpicsSignalRO, 'RingOverflows')
        trigger_mode = OCpt(EpicsSignal, 'TriggerMode')
        reset = OCpt(EpicsSignal, 'Reset')

    current_names = DDCpt(_current_fields('ch', 'CurrentName', range(1, 5),
                                          string=True))
    current_offsets = DDCpt(_current_fields('ch', 'CurrentOffset',
                                            range(1, 5)))
    current_offset_calcs = DDCpt(_current_fields('ch', 'ComputeCurrentOffset',
                                                 range(1, 5)))
    current_scales = DDCpt(_current_fields('ch', 'CurrentScale', range(1, 5)))

    position_offset_x = Cpt(EpicsSignal, 'PositionOffsetX')
    position_offset_y = Cpt(EpicsSignal, 'PositionOffsetY')

    position_offset_calc_x = Cpt(EpicsSignal, 'ComputePosOffsetX')
    position_offset_calc_y = Cpt(EpicsSignal, 'ComputePosOffsetY')

    position_scale_x = Cpt(EpicsSignal, 'PositionScaleX')
    position_scale_Y = Cpt(EpicsSignal, 'PositionScaleY')

    image = Cpt(ImagePlugin, 'image1:')
    current1 = Cpt(StatsPlugin, 'Current1:')
    current2 = Cpt(StatsPlugin, 'Current2:')
    current3 = Cpt(StatsPlugin, 'Current3:')
    current4 = Cpt(StatsPlugin, 'Current4:')

    sum_all = Cpt(StatsPlugin, 'SumAll:')

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.stage_sigs.update([('acquire', 0),  # if acquiring, stop
                                ('acquire_mode', 2)  # single mode
                                ])
        self._acquisition_signal = self.acquire

        for i in range(1, 5):
            current = getattr(self, 'current{}'.format(i))
            current.mean_value.kind = Kind.hinted


class NSLS_EM(QuadEM):
    ...


class TetrAMM(QuadEM):
    conf = Cpt(QuadEMPort, port_name='TetrAMM', kind='omitted')


class APS_EM(QuadEM):
    conf = Cpt(QuadEMPort, port_name='APS_EM', kind='omitted')



# vi: ts=4 sw=4 sts=4 expandtab
import logging
from collections import OrderedDict

from .errors import *  # noqa: F401, F403
from .epics_pvs import *  # noqa: F401, F403
from .paths import makedirs, make_dir_tree  # noqa: F401

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


def enum(**enums):
    '''Create an enum from the keyword arguments'''
    return type('Enum', (object,), enums)


class OrderedDefaultDict(OrderedDict):
    """
    a combination of defaultdict and OrderedDict

    source: http://stackoverflow.com/a/6190500/1221924
    """
    def __init__(self, default_factory=None, *a, **kw):
        if (default_factory is not None and not callable(default_factory)):
            raise TypeError('first argument must be callable')
        super().__init__(*a, **kw)
        self.default_factory = default_factory

    def __getitem__(self, key):
        try:
            return super().__getitem__(key)
        except KeyError:
            return self.__missing__(key)

    def __missing__(self, key):
        if self.default_factory is None:
            raise KeyError(key)
        self[key] = value = self.default_factory()
        return value

    def __reduce__(self):
        if self.default_factory is None:
            args = tuple()
        else:
            args = self.default_factory,
        return type(self), args, None, None, self.items()

    def copy(self):
        return self.__copy__()

    def __copy__(self):
        return type(self)(self.default_factory, self)

    def __deepcopy__(self, memo):
        import copy
        return type(self)(self.default_factory,
                          copy.deepcopy(self.items()))

    def __repr__(self):
        return '%s(%s, %s)' % (self.__class__.__name__, self.default_factory,
                               super().__repr__())


def doc_annotation_forwarder(base_klass):
    def wrapper(f):
        f_name = getattr(f, '__name__')
        base_func = getattr(base_klass, f_name)
        base_docs = getattr(base_func, '__doc__')
        base_annotation = getattr(base_func, '__annotations__')
        f.__doc__ = base_docs
        f.__annotations__ = base_annotation

        return f

    return wrapper


def _filtered_ip_ns():
    import IPython
    return {k: v
            for k, v in IPython.get_ipython().user_ns.items()
            if not k.startswith('_')}


def instances_from_namespace(classes, *, ns=None):
    '''Get instances of `classes` from the user namespace

    Parameters
    ----------
    classes : type, or sequence of types
        Passed directly to isinstance(), only instances of these classes
        will be returned.

    ns : Dict[str, Any], optional
       namespace to pull from, defaults to getting the
    '''
    if ns is None:
        ns = _filtered_ip_ns()
    return [val for val in ns.values()
            if isinstance(val, classes)]


def ducks_from_namespace(attrs, *, ns=None):
    '''Get instances that have all of attributes.

    "Ducks" is a reference to "duck-typing." If it looks like a duck....

    Parameters
    ----------
    attr : Union[str, Iterable[str]]
        name of attribute or list of names
    '''
    if isinstance(attrs, str):
        attrs = [attrs]
    if ns is None:
        ns = _filtered_ip_ns()
    return [val for val in ns.values()
            if all(hasattr(val, attr) for attr in attrs)]


def underscores_to_camel_case(underscores):
    'Convert abc_def_ghi to abcDefGhi'
    if '_' in underscores and underscores.strip('_') != '':
        return ''.join(s.capitalize() for s in underscores.split('_'))

    return underscores.capitalize()


def getattrs(obj, gen):
    'For each item in the generator `gen`, yields (attr, getattr(obj, attr))'
    for attr in gen:
        yield attr, getattr(obj, attr)


class DO_NOT_USE:
    "sentinel value"
    ...



import logging
import time
import copy
import pytest
import threading

from ophyd import get_cl
from ophyd.signal import (Signal, EpicsSignal, EpicsSignalRO, DerivedSignal)
from ophyd.utils import (ReadOnlyError, AlarmStatus, AlarmSeverity)
from ophyd.status import wait

logger = logging.getLogger(__name__)


@pytest.fixture(scope='function')
def ro_signal(cleanup, signal_test_ioc):
    sig = EpicsSignalRO(signal_test_ioc.pvs['pair_rbv'], name='pair_rbv')
    cleanup.add(sig)
    sig.wait_for_connection()
    return sig


@pytest.fixture(scope='function')
def bool_enum_signal(cleanup, signal_test_ioc):
    sig = EpicsSignal(signal_test_ioc.pvs['bool_enum'], name='bool_enum')
    cleanup.add(sig)
    sig.wait_for_connection()
    return sig


@pytest.fixture(scope='function')
def rw_signal(cleanup, signal_test_ioc):
    sig = EpicsSignal(signal_test_ioc.pvs['read_write'], name='read_write')
    cleanup.add(sig)
    sig.wait_for_connection()
    return sig


@pytest.fixture(scope='function')
def pair_signal(cleanup, signal_test_ioc):
    sig = EpicsSignal(read_pv=signal_test_ioc.pvs['pair_rbv'],
                      write_pv=signal_test_ioc.pvs['pair_set'],
                      name='pair')
    cleanup.add(sig)
    sig.wait_for_connection()
    return sig


@pytest.fixture(scope='function')
def motor_pair_signal(cleanup, motor):
    sig = EpicsSignal(write_pv=motor.user_setpoint.pvname,
                      read_pv=motor.user_readback.pvname)
    cleanup.add(sig)
    sig.wait_for_connection()
    return sig


@pytest.fixture(scope='function')
def set_severity_signal(cleanup, signal_test_ioc):
    sig = EpicsSignal(signal_test_ioc.pvs['set_severity'], name='set_severity')
    cleanup.add(sig)
    sig.wait_for_connection()
    return sig


@pytest.fixture(scope='function')
def alarm_status_signal(cleanup, signal_test_ioc):
    sig = EpicsSignal(signal_test_ioc.pvs['alarm_status'], name='alarm_status')
    cleanup.add(sig)
    sig.wait_for_connection()
    return sig


def test_signal_base():
    start_t = time.time()

    name = 'test'
    value = 10.0
    signal = Signal(name=name, value=value, timestamp=start_t)
    signal.wait_for_connection()

    assert signal.connected
    assert signal.name == name
    assert signal.value == value
    assert signal.get() == value
    assert signal.timestamp == start_t

    info = dict(called=False)

    def _sub_test(**kwargs):
        info['called'] = True
        info['kw'] = kwargs

    signal.subscribe(_sub_test, run=False,
                     event_type=signal.SUB_VALUE)
    assert not info['called']

    signal.value = value
    signal.clear_sub(_sub_test)

    signal.subscribe(_sub_test, run=False,
                     event_type=signal.SUB_VALUE)
    signal.clear_sub(_sub_test, event_type=signal.SUB_VALUE)

    kw = info['kw']
    assert 'value' in kw
    assert 'timestamp' in kw
    assert 'old_value' in kw

    assert kw['value'] == value
    assert kw['old_value'] == value
    assert kw['timestamp'] == signal.timestamp

    # readback callback for soft signal
    info = dict(called=False)
    signal.subscribe(_sub_test, event_type=Signal.SUB_VALUE,
                     run=False)
    assert not info['called']
    signal.put(value + 1)
    assert info['called']

    signal.clear_sub(_sub_test)
    kw = info['kw']

    assert 'value' in kw
    assert 'timestamp' in kw
    assert 'old_value' in kw

    assert kw['value'] == value + 1
    assert kw['old_value'] == value
    assert kw['timestamp'] == signal.timestamp

    signal.trigger()
    signal.read()
    signal.describe()
    signal.read_configuration()
    signal.describe_configuration()

    eval(repr(signal))


def test_signal_copy():
    start_t = time.time()

    name = 'test'
    value = 10.0
    signal = Signal(name=name, value=value, timestamp=start_t)
    sig_copy = copy.copy(signal)

    assert signal.name == sig_copy.name
    assert signal.value == sig_copy.value
    assert signal.get() == sig_copy.get()
    assert signal.timestamp == sig_copy.timestamp


def test_epicssignal_readonly(cleanup, signal_test_ioc):
    signal = EpicsSignalRO(signal_test_ioc.pvs['read_only'])
    cleanup.add(signal)
    signal.wait_for_connection()
    print('EpicsSignalRO.metadata=', signal.metadata)
    signal.get()

    assert not signal.write_access
    assert signal.read_access

    with pytest.raises(ReadOnlyError):
        signal.value = 10

    with pytest.raises(ReadOnlyError):
        signal.put(10)

    with pytest.raises(ReadOnlyError):
        signal.set(10)

    # vestigial, to be removed
    with pytest.raises(AttributeError):
        signal.setpoint_ts

    # vestigial, to be removed
    with pytest.raises(AttributeError):
        signal.setpoint

    signal.precision
    signal.timestamp
    signal.limits

    signal.read()
    signal.describe()
    signal.read_configuration()
    signal.describe_configuration()

    eval(repr(signal))
    time.sleep(0.2)


def test_epicssignal_readwrite_limits(pair_signal):
    signal = pair_signal
    signal.use_limits = True
    signal.check_value((signal.low_limit + signal.high_limit) / 2)

    with pytest.raises(ValueError):
        signal.check_value(None)

    with pytest.raises(ValueError):
        signal.check_value(signal.low_limit - 1)

    with pytest.raises(ValueError):
        signal.check_value(signal.high_limit + 1)


def test_epicssignal_readwrite(signal_test_ioc, pair_signal):
    pair_signal.use_limits = True
    signal = pair_signal

    assert signal.setpoint_pvname == signal_test_ioc.pvs['pair_set']
    assert signal.pvname == signal_test_ioc.pvs['pair_rbv']
    signal.get()

    time.sleep(0.2)

    value = 10
    signal.value = value
    signal.put(value)
    assert signal.setpoint == value
    signal.setpoint_ts

    signal.limits
    signal.precision
    signal.timestamp

    signal.read()
    signal.describe()
    signal.read_configuration()
    signal.describe_configuration()

    eval(repr(signal))
    time.sleep(0.2)


def test_epicssignal_waveform(cleanup, signal_test_ioc):
    called = False

    def update_cb(value=None, **kwargs):
        nonlocal called
        assert len(value) > 1
        called = True

    signal = EpicsSignal(signal_test_ioc.pvs['waveform'], string=True)
    cleanup.add(signal)
    signal.wait_for_connection()

    sub = signal.subscribe(update_cb, event_type=signal.SUB_VALUE)
    assert len(signal.get()) > 1
    # force the current thread to allow other threads to run to service
    # subscription
    time.sleep(0.2)
    assert called
    signal.unsubscribe(sub)


def test_no_connection(cleanup, signal_test_ioc):
    sig = EpicsSignal('does_not_connect')
    cleanup.add(sig)

    with pytest.raises(TimeoutError):
        sig.wait_for_connection()

    sig = EpicsSignal('does_not_connect')
    cleanup.add(sig)

    with pytest.raises(TimeoutError):
        sig.put(0.0)

    with pytest.raises(TimeoutError):
        sig.get()

    sig = EpicsSignal(signal_test_ioc.pvs['read_only'], write_pv='does_not_connect')
    cleanup.add(sig)
    with pytest.raises(TimeoutError):
        sig.wait_for_connection()


def test_enum_strs(bool_enum_signal):
    assert bool_enum_signal.enum_strs == ('Off', 'On')


def test_setpoint(rw_signal):
    rw_signal.get_setpoint()
    rw_signal.get_setpoint(as_string=True)


def test_epicssignalro():
    with pytest.raises(TypeError):
        # not in initializer parameters anymore
        EpicsSignalRO('test', write_pv='nope_sorry')


def test_describe(bool_enum_signal):
    sig = bool_enum_signal

    sig.put(1)
    desc = sig.describe()['bool_enum']
    assert desc['dtype'] == 'integer'
    assert desc['shape'] == []
    # assert 'precision' in desc
    assert desc['enum_strs'] == ('Off', 'On')
    assert 'upper_ctrl_limit' in desc
    assert 'lower_ctrl_limit' in desc

    sig = Signal(name='my_pv')
    sig.put('Off')
    desc = sig.describe()['my_pv']
    assert desc['dtype'] == 'string'
    assert desc['shape'] == []

    sig.put(3.14)
    desc = sig.describe()['my_pv']
    assert desc['dtype'] == 'number'
    assert desc['shape'] == []

    import numpy as np
    sig.put(np.array([1, ]))
    desc = sig.describe()['my_pv']
    assert desc['dtype'] == 'array'
    assert desc['shape'] == [1, ]


def test_set_method():
    sig = Signal(name='sig')

    st = sig.set(28)
    wait(st)
    assert st.done
    assert st.success
    assert sig.get() == 28


def test_soft_derived():
    timestamp = 1.0
    value = 'q'
    original = Signal(name='original', timestamp=timestamp, value=value)

    cb_values = []

    def callback(value=None, **kwargs):
        cb_values.append(value)

    derived = DerivedSignal(derived_from=original, name='derived')
    derived.subscribe(callback, event_type=derived.SUB_VALUE)

    assert derived.timestamp == timestamp
    assert derived.get() == value
    assert derived.timestamp == timestamp
    assert derived.describe()[derived.name]['derived_from'] == original.name
    assert derived.write_access == original.write_access
    assert derived.read_access == original.read_access

    new_value = 'r'
    derived.put(new_value)
    assert original.get() == new_value
    assert derived.get() == new_value
    assert derived.timestamp == original.timestamp
    assert derived.limits == original.limits

    copied = copy.copy(derived)
    assert copied.derived_from.value == original.value
    assert copied.derived_from.timestamp == original.timestamp
    assert copied.derived_from.name == original.name

    derived.put('s')
    assert cb_values == ['r', 's']

    called = []

    event = threading.Event()

    def meta_callback(*, connected, read_access, write_access, **kw):
        called.append(('meta', connected, read_access, write_access))
        event.set()

    derived.subscribe(meta_callback, event_type=derived.SUB_META, run=False)

    original._metadata['write_access'] = False
    original._run_subs(sub_type='meta', **original._metadata)

    event.wait(1)

    assert called == [('meta', True, True, False)]


def test_epics_signal_derived(ro_signal):
    assert ro_signal.connected
    assert ro_signal.read_access
    assert not ro_signal.write_access

    derived = DerivedSignal(derived_from=ro_signal, name='derived')
    derived.wait_for_connection()

    assert derived.connected
    assert derived.read_access
    assert not derived.write_access

    assert derived.timestamp == ro_signal.timestamp
    assert derived.get() == ro_signal.get()


@pytest.mark.motorsim
@pytest.mark.parametrize('put_complete', [True, False])
def test_epicssignal_set(motor_pair_signal, put_complete):
    sim_pv = motor_pair_signal
    sim_pv.put_complete = put_complete

    logging.getLogger('ophyd.signal').setLevel(logging.DEBUG)
    logging.getLogger('ophyd.utils.epics_pvs').setLevel(logging.DEBUG)
    print('tolerance=', sim_pv.tolerance)
    assert sim_pv.tolerance is not None

    start_pos = sim_pv.get()

    # move to +0.2 and check the status object
    target = sim_pv.get() + 0.2
    st = sim_pv.set(target, timeout=1, settle_time=0.001)
    wait(st, timeout=5)
    assert st.done
    assert st.success
    print('status 1', st)
    assert abs(target - sim_pv.get()) < 0.05

    # move back to -0.2, forcing a timeout with a low value
    target = sim_pv.get() - 0.2
    st = sim_pv.set(target, timeout=1e-6)
    time.sleep(0.5)
    print('status 2', st)
    assert st.done
    assert not st.success

    # keep the axis in position
    st = sim_pv.set(start_pos)
    wait(st, timeout=5)


statuses_and_severities = [
    (AlarmStatus.NO_ALARM, AlarmSeverity.NO_ALARM),
    (AlarmStatus.READ, AlarmSeverity.MINOR),
    (AlarmStatus.WRITE, AlarmSeverity.MAJOR),
    (AlarmStatus.HIHI, AlarmSeverity.INVALID),
    (AlarmStatus.NO_ALARM, AlarmSeverity.NO_ALARM),
]


@pytest.mark.parametrize('status, severity', statuses_and_severities)
def test_epicssignal_alarm_status(set_severity_signal, alarm_status_signal, pair_signal, status, severity):
    alarm_status_signal.put(status, wait=True)
    set_severity_signal.put(severity, wait=True)

    pair_signal.get()
    assert pair_signal.alarm_status == status
    assert pair_signal.alarm_severity == severity

    pair_signal.get_setpoint()
    assert pair_signal.setpoint_alarm_status == status
    assert pair_signal.setpoint_alarm_severity == severity


@pytest.mark.parametrize('status, severity', statuses_and_severities)
def test_epicssignalro_alarm_status(set_severity_signal, alarm_status_signal, ro_signal, status, severity):
    alarm_status_signal.put(status, wait=True)
    set_severity_signal.put(severity, wait=True)

    ro_signal.get()
    assert ro_signal.alarm_status == status
    assert ro_signal.alarm_severity == severity


def test_hints(cleanup, fake_motor_ioc):
    sig = EpicsSignalRO(fake_motor_ioc.pvs['setpoint'])
    cleanup.add(sig)
    assert sig.hints == {'fields': [sig.name]}


def test_epicssignal_sub_setpoint(cleanup, fake_motor_ioc):
    pvs = fake_motor_ioc.pvs
    pv = EpicsSignal(write_pv=pvs['setpoint'], read_pv=pvs['readback'],
                     name='pv')
    cleanup.add(pv)

    setpoint_called = []
    setpoint_meta_called = []

    def sub_setpoint(old_value, value, **kwargs):
        setpoint_called.append((old_value, value))

    def sub_setpoint_meta(timestamp, **kwargs):
        setpoint_meta_called.append(timestamp)

    pv.subscribe(sub_setpoint, event_type=pv.SUB_SETPOINT)
    pv.subscribe(sub_setpoint_meta, event_type=pv.SUB_SETPOINT_META)

    pv.wait_for_connection()

    pv.put(1, wait=True)
    pv.put(2, wait=True)
    time.sleep(1.0)

    assert len(setpoint_called) >= 3
    assert len(setpoint_meta_called) >= 3


def test_epicssignal_get_in_callback(fake_motor_ioc, cleanup):
    pvs = fake_motor_ioc.pvs
    sig = EpicsSignal(write_pv=pvs['setpoint'], read_pv=pvs['readback'],
                      name='motor')
    cleanup.add(sig)

    called = []

    def generic_sub(sub_type, **kwargs):
        called.append((sub_type, sig.get(), sig.get_setpoint()))

    for event_type in (sig.SUB_VALUE, sig.SUB_META,
                       sig.SUB_SETPOINT, sig.SUB_SETPOINT_META):
        sig.subscribe(generic_sub, event_type=event_type)

    sig.wait_for_connection()

    sig.put(1, wait=True)
    sig.put(2, wait=True)
    time.sleep(0.5)

    print(called)
    # Arbitrary threshold, but if @klauer screwed something up again, this will
    # blow up
    assert len(called) < 20
    print('total', len(called))
    sig.unsubscribe_all()


@pytest.mark.motorsim
@pytest.mark.parametrize('pvname, count',
                         [('sim:mtr1.RBV', 10),
                          ('sim:mtr2.RBV', 10),
                          ('sim:mtr1.RBV', 100),
                          ('sim:mtr2.RBV', 100),
                          ]
                         )
def test_epicssignal_pv_reuse(cleanup, pvname, count):
    signals = [EpicsSignal(pvname, name='sig')
               for i in range(count)]

    for sig in signals:
        cleanup.add(sig)
        sig.wait_for_connection()
        assert sig.connected
        assert sig.get() is not None

    if get_cl().name == 'pyepics':
        assert len(set(id(sig._read_pv) for sig in signals)) == 1



from unittest.mock import Mock
from ophyd import Device
from ophyd.status import (StatusBase, SubscriptionStatus, UseNewProperty,
                          MoveStatus)
import pytest


def _setup_state_and_cb():
    state = {}

    def cb():
        state['done'] = True
    return state, cb


def test_status_post():
    st = StatusBase()
    state, cb = _setup_state_and_cb()

    assert 'done' not in state
    st.add_callback(cb)
    assert 'done' not in state
    st._finished()
    assert 'done' in state
    assert state['done']


def test_status_legacy_finished_cb():
    st = StatusBase()
    state1, cb1 = _setup_state_and_cb()
    state2, cb2 = _setup_state_and_cb()

    # The old setter works for adding one callback.
    with pytest.warns(UserWarning):
        st.finished_cb = cb1
    # The new getter works.
    st.callbacks == set([cb1])
    # And the old getter works so long as there is just one callback set.
    with pytest.warns(UserWarning):
        assert st.finished_cb is cb1
    # As before, the old setter cannot be updated once set.
    with pytest.raises(UseNewProperty):
        st.finished_cb = cb2
    # But, using the new method, we can add another callback.
    st.add_callback(cb2)
    # Once we have two callbacks, the getter does not work.
    with pytest.raises(UseNewProperty):
        st.finished_cb
    # But the new getter does.
    st.callbacks == set([cb1, cb2])

    assert 'done' not in state1
    assert 'done' not in state2
    st._finished()
    assert 'done' in state1
    assert 'done' in state2


def test_status_pre():
    st = StatusBase()
    state, cb = _setup_state_and_cb()

    st._finished()

    assert 'done' not in state
    st.add_callback(cb)
    assert 'done' in state
    assert state['done']


def test_direct_done_setting():
    st = StatusBase()
    state, cb = _setup_state_and_cb()

    with pytest.raises(RuntimeError):
        st.done = True  # changing isn't allowed
    with pytest.warns(UserWarning):
        st.done = False  # but for now no-ops warn

    st._finished()

    with pytest.raises(RuntimeError):
        st.done = False  # changing isn't allowed
    with pytest.warns(UserWarning):
        st.done = True  # but for now no-ops warn


def test_subscription_status():
    # Arbitrary device
    d = Device("Tst:Prefix", name='test')
    # Mock callback
    m = Mock()

    # Full fake callback signature
    def cb(*args, done=False, **kwargs):
        # Run mock callback
        m()
        # Return finished or not
        return done

    status = SubscriptionStatus(d, cb, event_type=d.SUB_ACQ_DONE)

    # Run callbacks but do not mark as complete
    d._run_subs(sub_type=d.SUB_ACQ_DONE, done=False)
    assert m.called
    assert not status.done and not status.success

    # Run callbacks and mark as complete
    d._run_subs(sub_type=d.SUB_ACQ_DONE, done=True)
    assert status.done and status.success


def test_and():
    st1 = StatusBase()
    st2 = StatusBase()
    st3 = st1 & st2
    # make sure deep recursion works
    st4 = st1 & st3
    st5 = st3 & st4
    state1, cb1 = _setup_state_and_cb()
    state2, cb2 = _setup_state_and_cb()
    state3, cb3 = _setup_state_and_cb()
    state4, cb4 = _setup_state_and_cb()
    state5, cb5 = _setup_state_and_cb()
    st1.add_callback(cb1)
    st2.add_callback(cb2)
    st3.add_callback(cb3)
    st4.add_callback(cb4)
    st5.add_callback(cb5)
    st1._finished()
    assert 'done' in state1
    assert 'done' not in state2
    assert 'done' not in state3
    assert 'done' not in state4
    assert 'done' not in state5
    st2._finished()
    assert 'done' in state3
    assert 'done' in state4
    assert 'done' in state5
    assert st3.left is st1
    assert st3.right is st2
    assert st4.left is st1
    assert st4.right is st3
    assert st5.left is st3
    assert st5.right is st4


def test_notify_watchers():
    from ophyd.sim import hw
    hw = hw()
    mst = MoveStatus(hw.motor, 10)

    def callback(*args, **kwargs):
        ...

    mst.watch(callback)
    mst.target = 0
    mst.start_pos = 0
    mst._notify_watchers(0)



#!/usr/bin/env python
import os
import sys

from setuptools import (setup, find_packages)
import versioneer

# NOTE: This file must remain Python 2 compatible for the foreseeable future,
# to ensure that we error out properly for people with outdated setuptools
# and/or pip.
min_version = (3, 6)
if sys.version_info < min_version:
    error = """
ophyd does not support Python {0}.{1}.
Python {2}.{3} and above is required. Check your Python version like so:

python3 --version

This may be due to an out-of-date pip. Make sure you have pip >= 9.0.1.
Upgrade pip like so:

pip install --upgrade pip
""".format(*(sys.version_info[:2] + min_version))
    sys.exit(error)


here = os.path.abspath(os.path.dirname(__file__))

# Get the requirements from requirements.txt
with open(os.path.join(here, 'requirements.txt'), 'rt') as f:
    requirements = f.read().split()

# Get the long description from the README file
with open(os.path.join(here, 'README.md'), 'rt', encoding='utf-8') as f:
    long_description = f.read()

setup(name='ophyd',
      version=versioneer.get_version(),
      cmdclass=versioneer.get_cmdclass(),
      long_description=long_description,
      long_description_content_type='text/markdown',
      license='BSD',
      python_requires='>={}'.format('.'.join(str(n) for n in min_version)),
      install_requires=requirements,
      packages=find_packages(),
      entry_points={
          'databroker.handlers': [
              'NPY_SEQ = ophyd.sim:NumpySeqHandler',
          ]},
      classifiers=[
          "Development Status :: 5 - Production/Stable",
          "Programming Language :: Python :: 3.6",
          "Programming Language :: Python :: 3.7",
      ])
